
from qtoggleserver.core.typing import PortValue as CorePortValue

from .functions import function, Function


@function('AND')
class AndFunction(Function):
    MIN_ARGS = 2

    def eval(self) -> CorePortValue:
        r = True
        for e in self.eval_args():
            r = r and bool(e)

        return int(r)


@function('OR')
class OrFunction(Function):
    MIN_ARGS = 2

    def eval(self) -> CorePortValue:
        r = False
        for e in self.eval_args():
            r = r or bool(e)

        return int(r)


@function('NOT')
class NotFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def eval(self) -> CorePortValue:
        return int(not bool(self.eval_args()[0]))


@function('XOR')
class XOrFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        e1 = bool(eval_args[0])
        e2 = bool(eval_args[1])

        return int(e1 and not e2 or e2 and not e1)




from typing import Any, Dict, List, Union


PortValue = Union[int, float, bool]
NullablePortValue = Union[int, float, bool, None]
PortValueChoices = List[Dict[str, Union[str, int, float]]]

Attribute = Union[int, float, bool, str, List[dict]]
Attributes = Dict[str, Attribute]

AttributeDefinition = Dict[str, Any]
AttributeDefinitions = Dict[str, AttributeDefinition]

GenericJSONDict = Dict[str, Any]




import logging
import re

from typing import Any, Dict, Iterable, List, Optional, Union

from bson import ObjectId
from pymongo import MongoClient
from pymongo.database import Database

from qtoggleserver.persist import BaseDriver, Id, Record


logger = logging.getLogger(__name__)

_OBJECT_ID_RE = re.compile('^[0-9a-f]{24}$')
DEFAULT_DB = 'qtoggleserver'


class MongoDriver(BaseDriver):
    def __init__(self, host: str = '127.0.0.1', port: str = 27017, db: str = DEFAULT_DB, **kwargs) -> None:
        logger.debug('connecting to %s:%s/%s', host, port, db)

        self._client: MongoClient = MongoClient(host, port, serverSelectionTimeoutMS=200)
        self._db: Database = self._client[db]

    def query(
        self,
        collection: str,
        fields: Optional[List[str]],
        filt: Dict[str, Any],
        limit: Optional[int]
    ) -> Iterable[Record]:

        if fields:
            fields = dict((f, 1) for f in fields)
            if 'id' in fields:
                fields['_id'] = 1

            else:
                fields['_id'] = 0

        if 'id' in filt:
            filt = dict(filt)
            filt['_id'] = self._id_to_db(filt.pop('id'))

        q = self._db[collection].find(filt, fields)
        if limit is not None:
            q = q.limit(limit)

        return self._query_gen_wrapper(q)

    def insert(self, collection: str, record: Record) -> Id:
        if 'id' in record:
            record = dict(record)
            record['_id'] = self._id_to_db(record.pop('id'))

        return self._db[collection].insert_one(record).inserted_id

    def update(self, collection: str, record_part: Record, filt: Dict[str, Any]) -> int:
        if 'id' in record_part:
            record_part = dict(record_part)
            record_part['_id'] = self._id_to_db(record_part.pop('id'))

        if 'id' in filt:
            filt = dict(filt)
            filt['_id'] = self._id_to_db(filt.pop('id'))

        return self._db[collection].update_many(filt, {'$set': record_part}, upsert=False).modified_count

    def replace(self, collection: str, _id: Id, record: Record, upsert: bool) -> bool:
        record = dict(record)
        record.pop('id', None)
        record['_id'] = self._id_to_db(_id)

        return bool(self._db[collection].replace_one({'_id': record['_id']}, record, upsert=upsert).modified_count)

    def remove(self, collection: str, filt: Dict[str, Any]) -> int:
        if 'id' in filt:
            filt = dict(filt)
            filt['_id'] = self._id_to_db(filt.pop('id'))

        return self._db[collection].delete_many(filt).deleted_count

    def cleanup(self) -> None:
        logger.debug('disconnecting mongo client')

        self._client.close()

    @classmethod
    def _query_gen_wrapper(cls, q: Iterable[Record]) -> Iterable[Record]:
        for r in q:
            if '_id' in r:
                r['id'] = cls._id_from_db(r.pop('_id'))

            yield r

    @staticmethod
    def _id_to_db(_id: str) -> Union[str, ObjectId]:
        if isinstance(_id, str) and _OBJECT_ID_RE.match(_id):
            return ObjectId(_id)

        else:
            return _id

    @staticmethod
    def _id_from_db(_id: Union[str, ObjectId]) -> str:
        if isinstance(_id, ObjectId):
            return str(_id)

        else:
            return _id




import time

from typing import Set

from qtoggleserver.core.typing import PortValue as CorePortValue

from .functions import function, Function


@function('TIME')
class TimeFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self) -> Set[str]:
        return {'time'}

    def eval(self) -> CorePortValue:
        return int(time.time())


@function('TIMEMS')
class TimeMSFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self) -> Set[str]:
        return {'time_ms'}

    def eval(self) -> CorePortValue:
        return int(time.time() * 1000)




from __future__ import annotations

import abc

from qtoggleserver.core import ports as core_ports

from .peripheral import Peripheral


class PeripheralPort(core_ports.Port, metaclass=abc.ABCMeta):
    ID = 'port'

    def __init__(self, peripheral: Peripheral) -> None:
        self._peripheral: Peripheral = peripheral

        _id = self.make_id()
        if self._peripheral.get_name():
            _id = f'{self._peripheral.get_name()}.{_id}'

        super().__init__(_id)

    def make_id(self) -> str:
        return self.ID

    def get_peripheral(self) -> Peripheral:
        return self._peripheral

    async def handle_enable(self) -> None:
        if not self._peripheral.is_enabled():
            await self._peripheral.enable()

    async def handle_disable(self) -> None:
        if self._peripheral.is_enabled():
            await self._peripheral.check_disabled(self)

    async def attr_is_online(self) -> bool:
        if not self.is_enabled():
            return False

        return self._peripheral.is_online()




import abc
import datetime
import logging

from typing import Dict, Optional, Tuple

from jinja2 import Environment, Template

from qtoggleserver.core import device
from qtoggleserver.core import events as core_events
from qtoggleserver.core import main as core_main
from qtoggleserver.core import ports as core_ports
from qtoggleserver.conf import settings
from qtoggleserver.core.typing import Attribute, Attributes, NullablePortValue
from qtoggleserver.slaves import devices as slaves_devices

from .filtereventhandler import FilterEventHandler


logger = logging.getLogger(__name__)


class TemplateNotificationsHandler(FilterEventHandler, metaclass=abc.ABCMeta):
    DEFAULT_TEMPLATES = {
        'value-change': {
            'title': '{{port.get_display_name()}} is {{new_value}}{{attrs["unit"]}}',
            'body': 'Port {{port.get_display_name()}} was {{old_value}}{{attrs["unit"]}} '
                    'and is now {{new_value}}{{attrs["unit"]}}.'
        },
        'port-update': {
            'title': '{{port.get_display_name()}} has been updated',
            'body': 'Port {{port.get_display_name()}} attributes have been updated.'
        },
        'port-add': {
            'title': '{{port.get_display_name()}} has been added',
            'body': None
        },
        'port-remove': {
            'title': '{{port.get_display_name()}} has been removed',
            'body': None
        },
        'device-update': {
            'title': '{{device.get_display_name()}} has been updated',
            'body': 'Device {{device.get_display_name()}} attributes have been updated.'
        },
        'full-update': {
            'title': '{{device.get_display_name()}} has been updated',
            'body': 'Device {{device.get_display_name()}} has been fully updated.'
        },
        'slave-device-update': {
            'title': '{{slave.get_display_name()}} has been updated',
            'body': 'Device {{slave.get_display_name()}} attributes have been updated.'
        },
        'slave-device-add': {
            'title': '{{slave.get_display_name()}} has been added',
            'body': None
        },
        'slave-device-remove': {
            'title': '{{slave.get_display_name()}} has been removed',
            'body': None
        },
    }

    logger = logger

    def __init__(
        self,
        *,
        template: Optional[Dict[str, str]] = None,
        templates: Optional[Dict[str, Dict[str, str]]] = None,
        skip_startup: bool = True,
        filter: dict = None,
        name: Optional[str] = None
    ) -> None:

        self._skip_startup: bool = skip_startup

        # "template" has the highest precedence; then comes "templates" and then comes "DEFAULT_TEMPLATES"
        if template is not None:
            templates = {k: template for k in self.DEFAULT_TEMPLATES.keys()}
        if templates is None:
            templates = self.DEFAULT_TEMPLATES

        # Convert template strings to jinja2 templates
        self._j2env: Environment = Environment()
        self._templates: Dict[str, Dict[str, Template]] = {}
        for _type, ts in templates.items():
            # Ensure values in templates are dicts themselves
            if isinstance(ts, str):
                ts = {'title': ts}

            self._templates[_type] = {}
            for k, t in ts.items():
                self._templates[_type][k] = self._j2env.from_string(t) if t is not None else None

        super().__init__(name=name, filter=filter)

    def render(self, event_type: str, context: dict) -> Dict[str, str]:
        template = self._templates[event_type]

        return {k: t.render(context) if t is not None else None for k, t in template.items()}

    @staticmethod
    def get_common_context(event: core_events.Event) -> dict:
        timestamp = event.get_timestamp()
        moment = datetime.datetime.fromtimestamp(timestamp)

        return {
            'event': event,
            'type': event.get_type(),
            'timestamp': int(timestamp),
            'moment': moment,
            'display_moment': moment.strftime('%c'),
            'public_url': settings.public_url
        }

    @abc.abstractmethod
    async def push_message(self, event: core_events.Event, title: str, body: str, **kwargs) -> None:
        raise NotImplementedError()

    async def push_template_message(self, event: core_events.Event, context: dict) -> None:
        template = self.render(event.get_type(), context)

        await self.push_message(event, **template)

    async def accepts(self, *args, **kwargs) -> bool:
        # Skip notifications during startup
        if self._skip_startup:
            if not core_main.is_ready():
                return False

            if isinstance(self._skip_startup, (int, float)) and core_main.uptime() < self._skip_startup:
                return False

        return await super().accepts(*args, **kwargs)

    async def on_value_change(
        self,
        event: core_events.Event,
        port: core_ports.BasePort,
        old_value: NullablePortValue,
        new_value: NullablePortValue,
        attrs: Attributes
    ) -> None:

        context = self.get_common_context(event)
        context.update({
            'port': port,
            'old_value': old_value,
            'new_value': new_value,
            'attrs': attrs
        })

        await self.push_template_message(event, context)

    async def on_port_update(
        self,
        event: core_events.Event,
        port: core_ports.BasePort,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> None:

        context = self.get_common_context(event)
        context.update({
            'port': port,
            'old_attrs': old_attrs,
            'new_attrs': new_attrs,
            'changed_attrs': changed_attrs,
            'added_attrs': added_attrs,
            'removed_attrs': removed_attrs,
            'value': port.get_value()
        })

        await self.push_template_message(event, context)

    async def on_port_add(self, event: core_events.Event, port: core_ports.BasePort, attrs: Attributes) -> None:
        context = self.get_common_context(event)
        context.update({
            'port': port,
            'attrs': attrs,
            'value': port.get_value()
        })

        await self.push_template_message(event, context)

    async def on_port_remove(self, event: core_events.Event, port: core_ports.BasePort, attrs: Attributes) -> None:
        context = self.get_common_context(event)
        context.update({
            'port': port,
            'attrs': attrs,
            'value': port.get_value()
        })

        await self.push_template_message(event, context)

    async def on_device_update(
        self,
        event: core_events.Event,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> None:

        context = self.get_common_context(event)
        context.update({
            'device': device,
            'old_attrs': old_attrs,
            'new_attrs': new_attrs,
            'changed_attrs': changed_attrs,
            'added_attrs': added_attrs,
            'removed_attrs': removed_attrs,
        })

        await self.push_template_message(event, context)

    async def on_full_update(self, event: core_events.Event) -> None:
        context = self.get_common_context(event)

        await self.push_template_message(event, context)

    async def on_slave_device_update(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> None:

        context = self.get_common_context(event)
        context.update({
            'slave': slave,
            'old_attrs': old_attrs,
            'new_attrs': new_attrs,
            'changed_attrs': changed_attrs,
            'added_attrs': added_attrs,
            'removed_attrs': removed_attrs,
        })

        await self.push_template_message(event, context)

    async def on_slave_device_add(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave,
        attrs: Attributes
    ) -> None:

        context = self.get_common_context(event)
        context.update({
            'slave': slave,
            'attrs': attrs
        })

        await self.push_template_message(event, context)

    async def on_slave_device_remove(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave,
        attrs: Attributes
    ) -> None:

        context = self.get_common_context(event)
        context.update({
            'slave': slave,
            'attrs': attrs
        })

        await self.push_template_message(event, context)




import asyncio
import inspect
import logging

from typing import Any, Callable, Dict, Optional

from tornado.iostream import StreamClosedError
from tornado.web import RequestHandler, HTTPError

from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.core.api import funcs as core_api_funcs
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.slaves.api import funcs as slaves_api_funcs
from qtoggleserver.ui.api import funcs as ui_api_funcs
from qtoggleserver.utils import json as json_utils


logger = logging.getLogger(__name__)


class NoSuchFunction(HTTPError):
    def __init__(self) -> None:
        super().__init__(404, 'no-such-function')


class BaseHandler(RequestHandler):
    _UNDEFINED = {}

    def __init__(self, *args, **kwargs) -> None:
        self._json: Any = self._UNDEFINED
        self._response_body: str = ''
        self._response_body_json: Any = None

        RequestHandler.__init__(self, *args, **kwargs)

    def get_request_json(self) -> Any:
        if self._json is self._UNDEFINED:
            try:
                self._json = json_utils.loads(self.request.body)

            except ValueError as e:
                logger.error('could not decode json from request body: %s', e)

                raise core_api.APIError(400, 'malformed-body') from e

        return self._json

    def finish(self, chunk: Optional[str] = None) -> asyncio.Future:
        self._response_body = chunk

        return super().finish(chunk)

    def finish_json(self, data: Any) -> asyncio.Future:
        self._response_body_json = data

        data = json_utils.dumps(data)
        data += '\n'

        self.set_header('Content-Type', 'application/json; charset=utf-8')
        return self.finish(data)

    def get_response_body(self) -> str:
        return self._response_body

    def get_response_body_json(self) -> Any:
        return self._response_body_json

    def get_response_headers(self) -> Dict[str, str]:
        return dict(self._headers.get_all())

    def get(self, **kwargs) -> None:
        raise NoSuchFunction()

    head = post = delete = patch = put = options = get

    def _handle_request_exception(self, exception: Exception) -> None:
        try:
            if isinstance(exception, HTTPError):
                logger.error('%s %s: %s', self.request.method, self.request.uri, exception)
                self.set_status(exception.status_code)
                self.finish_json({'error': (exception.log_message or
                                            getattr(exception, 'reason', None) or str(exception))})

            else:
                logger.error(str(exception), exc_info=True)
                self.set_status(500)
                self.finish_json({'error': 'internal server error'})

        except RuntimeError:
            pass  # Nevermind

    def data_received(self, chunk: bytes) -> None:
        pass


class NoSuchFunctionHandler(BaseHandler):
    pass


class APIHandler(BaseHandler):
    AUTH_ENABLED = True

    def __init__(self, *args, **kwargs) -> None:
        self.access_level: int = core_api.ACCESS_LEVEL_NONE
        self.username: Optional[str] = None

        BaseHandler.__init__(self, *args, **kwargs)

    def prepare(self) -> None:
        # Disable cache
        self.set_header('Cache-Control', 'no-cache, no-store, must-revalidate, max-age=0')

        if not self.AUTH_ENABLED:
            return

        # Parse auth header
        auth = self.request.headers.get('Authorization')
        if auth:
            try:
                usr = core_api_auth.parse_auth_header(
                    auth,
                    core_api_auth.ORIGIN_CONSUMER,
                    core_api_auth.consumer_password_hash_func
                )

            except core_api_auth.AuthError as e:
                logger.warning(str(e))
                return

        else:
            if core_device_attrs.admin_password_hash == core_device_attrs.EMPTY_PASSWORD_HASH:
                logger.debug('authenticating request as admin due to empty admin password')
                usr = 'admin'

            else:
                logger.warning('missing authorization header')
                return

        self.access_level = core_api.ACCESS_LEVEL_MAPPING[usr]
        self.username = usr

        logger.debug(
            'granted access level %s (username=%s)',
            core_api.ACCESS_LEVEL_MAPPING[self.access_level],
            self.username
        )

    async def call_api_func(self, func: Callable, default_status: int = 200, **kwargs) -> None:
        try:
            if self.request.method in ('POST', 'PATCH', 'PUT'):
                kwargs['params'] = self.get_request_json()

            response = func(self, **kwargs)
            if inspect.isawaitable(response):
                response = await response

            self.set_status(default_status)
            if response is not None or default_status == 200:
                await self.finish_json(response)

            else:
                await self.finish()

        except Exception as e:
            await self._handle_api_call_exception(func, kwargs, e)

    async def _handle_api_call_exception(self, func: Callable, kwargs: dict, error: Exception) -> None:
        kwargs = dict(kwargs)
        params = kwargs.pop('params', None)
        args = json_utils.dumps(kwargs)
        body = params and json_utils.dumps(params) or '{}'

        if isinstance(error, core_responses.HTTPError):
            error = core_api.APIError.from_http_error(error)

        if isinstance(error, core_api.APIError):
            logger.error('api call %s failed: %s (args=%s, body=%s)', func.__name__, error, args, body)

            self.set_status(error.status)
            if not self._finished:  # Avoid finishing an already finished request
                await self.finish_json(error.to_json())

        elif isinstance(error, StreamClosedError) and func.__name__ == 'get_listen':
            logger.debug('api call get_listen could not complete: stream closed')

        else:
            logger.error('api call %s failed: %s (args=%s, body=%s)', func.__name__, error, args, body, exc_info=True)

            self.set_status(500)
            if not self._finished:  # Avoid finishing an already finished request
                await self.finish_json({'error': str(error)})


class DeviceHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(core_api_funcs.get_device)

    async def patch(self) -> None:
        await self.call_api_func(core_api_funcs.patch_device, default_status=204)


class ResetHandler(APIHandler):
    async def post(self) -> None:
        await self.call_api_func(core_api_funcs.post_reset, default_status=204)


class FirmwareHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(core_api_funcs.get_firmware)

    async def patch(self) -> None:
        await self.call_api_func(core_api_funcs.patch_firmware, default_status=204)


class AccessHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(core_api_funcs.get_access, access_level=self.access_level)


class SlaveDevicesHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(slaves_api_funcs.get_slave_devices)

    async def post(self) -> None:
        await self.call_api_func(slaves_api_funcs.post_slave_devices, default_status=201)


class SlaveDeviceHandler(APIHandler):
    async def patch(self, name: str) -> None:
        await self.call_api_func(slaves_api_funcs.patch_slave_device, name=name, default_status=204)

    async def delete(self, name: str) -> None:
        await self.call_api_func(slaves_api_funcs.delete_slave_device, name=name, default_status=204)


class SlaveDeviceEventsHandler(APIHandler):
    AUTH_ENABLED = False  # We'll take care of the authentication inside API call functions

    async def post(self, name: str) -> None:
        await self.call_api_func(slaves_api_funcs.post_slave_device_events, name=name, default_status=204)


class SlaveDeviceForwardHandler(APIHandler):
    async def get(self, name: str, path: str) -> None:
        await self.call_api_func(
            slaves_api_funcs.slave_device_forward,
            name=name,
            method=self.request.method,
            path=path
        )

    post = patch = delete = get


class DiscoveredHandler(APIHandler):
    async def get(self) -> None:
        timeout = self.get_argument('timeout', None)
        if timeout:
            timeout = int(timeout)

        await self.call_api_func(slaves_api_funcs.get_discovered, timeout=timeout)

    async def delete(self) -> None:
        await self.call_api_func(slaves_api_funcs.delete_discovered, default_status=204)


class DiscoveredDeviceHandler(APIHandler):
    async def patch(self, name: str) -> None:
        await self.call_api_func(slaves_api_funcs.patch_discovered_device, name=name)


class PortsHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(core_api_funcs.get_ports)

    async def post(self) -> None:
        if not settings.core.virtual_ports:
            raise NoSuchFunction()

        await self.call_api_func(core_api_funcs.post_ports, default_status=201)


class PortHandler(APIHandler):
    async def delete(self, port_id: str) -> None:
        if not settings.core.virtual_ports:
            raise NoSuchFunction()

        await self.call_api_func(core_api_funcs.delete_port, port_id=port_id, default_status=204)

    async def patch(self, port_id: str) -> None:
        await self.call_api_func(core_api_funcs.patch_port, port_id=port_id, default_status=204)


class PortValueHandler(APIHandler):
    async def get(self, port_id: str) -> None:
        await self.call_api_func(core_api_funcs.get_port_value, port_id=port_id)

    async def patch(self, port_id: str) -> None:
        await self.call_api_func(core_api_funcs.patch_port_value, port_id=port_id, default_status=204)


class PortSequenceHandler(APIHandler):
    async def patch(self, port_id: str) -> None:
        if not settings.core.sequences_support:
            raise NoSuchFunction()

        await self.call_api_func(core_api_funcs.patch_port_sequence, port_id=port_id, default_status=204)


class WebhooksHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(core_api_funcs.get_webhooks)

    async def patch(self) -> None:
        await self.call_api_func(core_api_funcs.patch_webhooks, default_status=204)


class ListenHandler(APIHandler):
    async def get(self) -> None:
        session_id = self.get_argument('session_id', None)
        timeout = self.get_argument('timeout', None)
        if timeout:
            timeout = int(timeout)

        await self.call_api_func(
            core_api_funcs.get_listen,
            session_id=session_id,
            timeout=timeout,
            access_level=self.access_level
        )


class ReverseHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(core_api_funcs.get_reverse)

    async def patch(self) -> None:
        await self.call_api_func(core_api_funcs.patch_reverse, default_status=204)


class DashboardPanelsHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(ui_api_funcs.get_panels)

    async def put(self) -> None:
        await self.call_api_func(ui_api_funcs.put_panels, default_status=204)


class PrefsHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(ui_api_funcs.get_prefs)

    async def put(self) -> None:
        await self.call_api_func(ui_api_funcs.put_prefs, default_status=204)




from qtoggleserver.core.typing import PortValue as CorePortValue

from .functions import function, Function


@function('IF')
class IfFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        if eval_args[0]:
            return eval_args[1]

        else:
            return eval_args[2]


@function('EQ')
class EqFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        return int(eval_args[0] == eval_args[1])


@function('GT')
class GTFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        return int(eval_args[0] > eval_args[1])


@function('GTE')
class GTEFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        return int(eval_args[0] >= eval_args[1])


@function('LT')
class LTFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        return int(eval_args[0] < eval_args[1])


@function('LTE')
class LTEFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        return int(eval_args[0] <= eval_args[1])




import logging

from qtoggleserver import persist

from . import devices as slaves_devices
from . import discover


logger = logging.getLogger(__name__)


def reset() -> None:
    logger.debug('clearing slaves persisted data')
    persist.remove('slaves')
    persist.remove('slave_ports')


async def init() -> None:
    logger.debug('loading devices')
    await slaves_devices.load()

    logger.debug('initializing discover mechanism')
    await discover.init()


async def cleanup() -> None:
    logger.debug('cleaning up discover mechanism')
    await discover.cleanup()

    logger.debug('cleaning up devices')
    await slaves_devices.cleanup()




import logging

import qui

from qtoggleserver import version
from qtoggleserver.conf import settings
from qtoggleserver.slaves.discover import is_enabled as is_discover_enabled


logger = logging.getLogger(__name__)


async def init() -> None:
    if settings.frontend.enabled:
        logger.debug('initializing QUI')

        qui.configure(
            name='qtoggleserver',
            display_name='qToggleServer',
            description='qToggleServer',
            version=version.VERSION,
            debug=settings.frontend.debug,
            static_url=settings.frontend.static_url,
            extra_context=dict(
                slaves_enabled=settings.slaves.enabled,
                discover_enabled=is_discover_enabled()
            )
        )


async def cleanup() -> None:
    pass




from .base import Event, Handler
from .device import DeviceEvent, DeviceUpdate, FullUpdate
from .handlers import handle_event, register_handler
from .handlers import init as init_handlers, cleanup as cleanup_handlers
from .port import PortEvent, PortAdd, PortRemove, PortUpdate, ValueChange


async def init() -> None:
    await init_handlers()


async def cleanup() -> None:
    await cleanup_handlers()




from qtoggleserver.core.typing import PortValue as CorePortValue

from .functions import function, Function


@function('MIN')
class MinFunction(Function):
    MIN_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        m = eval_args[0]
        for e in eval_args[1:]:
            if e < m:
                m = e

        return m


@function('MAX')
class MaxFunction(Function):
    MIN_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        m = eval_args[0]
        for e in eval_args[1:]:
            if e > m:
                m = e

        return m


@function('AVG')
class AvgFunction(Function):
    MIN_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        return sum(eval_args) / len(eval_args)




import abc
import logging

from typing import Dict, List, Optional, Set, Tuple, Union

from qtoggleserver.core import events as core_events
from qtoggleserver.core import expressions as core_expressions
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import Attribute, Attributes, NullablePortValue
from qtoggleserver.slaves import devices as slaves_devices
from qtoggleserver.slaves import events as slaves_events


logger = logging.getLogger(__package__)


class FilterEventHandler(core_events.Handler, metaclass=abc.ABCMeta):
    logger = logger

    def __init__(self, *, filter: dict = None, name: Optional[str] = None) -> None:
        self._filter: dict = filter or {}
        self._filter_prepared: bool = False

        self._filter_event_types: Optional[Set[str]] = None

        self._filter_device_attrs: Attributes = {}
        self._filter_device_attr_transitions: Dict[str, Tuple[Attribute, Attribute]] = {}
        self._filter_device_attr_names: Set[str] = set()

        self._filter_port_value: Optional[Union[List[NullablePortValue]], core_expressions.Expression] = None
        self._filter_port_value_transition: Optional[Tuple[NullablePortValue, NullablePortValue]] = None
        self._filter_port_attrs: Union[Attributes, Dict[str, List[Attribute]]] = {}
        self._filter_port_attr_transitions: Dict[str, Tuple[Attribute, Attribute]] = {}
        self._filter_port_attr_names: Set[str] = set()

        self._filter_slave_attrs: Union[Attributes, Dict[str, List[Attribute]]] = {}
        self._filter_slave_attr_transitions: Dict[str, Tuple[Attribute, Attribute]] = {}
        self._filter_slave_attr_names: Set[str] = set()

        # Maintain an internal "last" state for all objects, so we can detect changes in attributes and values
        self._device_attrs: Union[Attributes, Dict[str, List[Attribute]]] = {}
        self._port_values: Dict[str, NullablePortValue] = {}
        self._port_attrs: Dict[str, Attributes] = {}
        self._slave_attrs: Dict[str, Attributes] = {}

        super().__init__(name=name)

    def _prepare_filter(self) -> None:
        event_types = self._filter.get('type')
        if event_types:
            if not isinstance(event_types, list):
                event_types = (event_types,)
            self._filter_event_types = set(event_types)

        port_value = self._filter.get('port_value')
        if port_value is not None:
            if isinstance(port_value, str):  # An expression
                try:
                    self.debug('using value expression "%s"', port_value)
                    self._filter_port_value = core_expressions.parse(self_port_id=None, sexpression=port_value)

                except core_expressions.ExpressionParseError as e:
                    self.error('failed to parse port expression "%s": %s', port_value, e)

                    raise

            else:
                if not isinstance(port_value, list):
                    port_value = (port_value,)

                self._filter_port_value = port_value

        self._filter_port_value_transition = self._filter.get('port_value_transition')

        self._filter_device_attrs = {
            n[7:]: v for n, v in self._filter.items()
            if n.startswith('device_') and not n.endswith('_transition')
        }

        self._filter_device_attr_transitions = {
            n[7:-11]: v for n, v in self._filter.items()
            if n.startswith('device_') and n.endswith('_transition')
        }

        self._filter_port_attrs = {
            n[5:]: v for n, v in self._filter.items()
            if n.startswith('port_') and n != 'port_value' and not n.endswith('_transition')
        }

        self._filter_port_attr_transitions = {
            n[5:-11]: v for n, v in self._filter.items()
            if n.startswith('port_') and n != 'port_value_transition' and n.endswith('_transition')
        }

        self._filter_slave_attrs = {
            n[6:]: v for n, v in self._filter.items()
            if n.startswith('slave_') and not n.endswith('_transition')
        }

        self._filter_slave_attr_transitions = {
            n[6:-11]: v for n, v in self._filter.items()
            if n.startswith('slave_') and n.endswith('_transition')
        }

        self._filter_device_attr_names.update(self._filter_device_attrs.keys())
        self._filter_device_attr_names.update(self._filter_device_attr_transitions.keys())

        self._filter_port_attr_names.update(self._filter_port_attrs.keys())
        self._filter_port_attr_names.update(self._filter_port_attr_transitions.keys())

        self._filter_slave_attr_names.update(self._filter_slave_attrs.keys())
        self._filter_slave_attr_names.update(self._filter_slave_attr_transitions.keys())

        self._filter_prepared = True
        self.debug('filter prepared')

    @staticmethod
    def _make_changed_added_removed(old_attrs: Attributes, new_attrs: Attributes) -> Tuple[
        Dict[str, Tuple[Attribute, Attribute]],
        Attributes,
        Attributes
    ]:

        changed_attrs = {}
        added_attrs = {}
        removed_attrs = {}

        all_attr_names = set(old_attrs) | set(new_attrs)
        for n in all_attr_names:
            old_v = old_attrs.get(n)
            new_v = new_attrs.get(n)
            if old_v == new_v:
                continue

            if old_v is None:
                if new_v is not None:
                    added_attrs[n] = new_v

            elif new_v is None:
                removed_attrs[n] = old_v

            else:
                changed_attrs[n] = (old_v, new_v)

        return changed_attrs, added_attrs, removed_attrs

    async def _update_from_event(self, event: core_events.Event) -> Tuple[
        Tuple[NullablePortValue, NullablePortValue],
        Attributes,
        Attributes,
        Dict[str, Tuple[Attribute, Attribute]],
        Attributes,
        Attributes
    ]:

        value_pair = (None, None)
        old_attrs = {}
        new_attrs = {}
        changed_attrs = {}
        added_attrs = {}
        removed_attrs = {}

        if isinstance(event, core_events.PortEvent):
            port = event.get_port()

            old_attrs = self._port_attrs.get(port.get_id(), {})
            new_attrs = await port.get_attrs()

            old_value = self._port_values.get(port.get_id())
            new_value = port.get_value()
            value_pair = (old_value, new_value)

            if isinstance(event, (core_events.PortAdd, core_events.PortUpdate)):
                changed_attrs, added_attrs, removed_attrs = self._make_changed_added_removed(old_attrs, new_attrs)
                self._port_values[port.get_id()] = new_value
                self._port_attrs[port.get_id()] = new_attrs

            elif isinstance(event, core_events.PortRemove):
                self._port_values.pop(port.get_id(), None)
                removed_attrs = self._port_attrs.pop(port.get_id(), {})

            elif isinstance(event, core_events.ValueChange):
                self._port_values[port.get_id()] = new_value

        elif isinstance(event, core_events.DeviceEvent):
            old_attrs = self._device_attrs
            new_attrs = event.get_attrs()

            changed_attrs, added_attrs, removed_attrs = self._make_changed_added_removed(old_attrs, new_attrs)
            self._device_attrs = new_attrs

        elif isinstance(event, slaves_events.SlaveDeviceEvent):
            slave = event.get_slave()
            slave_json = slave.to_json()

            # Flatten slave master properties and attributes
            old_attrs = self._slave_attrs.get(slave.get_name(), {})
            new_attrs = dict(slave_json, **slave_json.pop('attrs'))

            if isinstance(event, (slaves_events.SlaveDeviceAdd, slaves_events.SlaveDeviceUpdate)):
                changed_attrs, added_attrs, removed_attrs = self._make_changed_added_removed(old_attrs, new_attrs)
                self._slave_attrs[slave.get_name()] = new_attrs

            elif isinstance(event, slaves_events.SlaveDeviceRemove):
                removed_attrs = self._slave_attrs.pop(slave.get_name(), {})

        return value_pair, old_attrs, new_attrs, changed_attrs, added_attrs, removed_attrs

    @staticmethod
    def _accepts_attrs(
        attr_names: Set[str],
        filter_attrs: Attributes,
        filter_attr_transitions: Dict[str, Tuple[Attribute, Attribute]],
        old_attrs: Attributes,
        new_attrs: Attributes
    ) -> bool:

        for name in attr_names:
            old_value = old_attrs.get(name)
            new_value = new_attrs.get(name)

            filter_transition = filter_attr_transitions.get(name)
            if filter_transition is not None:
                old_filter_value, new_filter_value = filter_transition
                if ((old_filter_value != old_value and old_filter_value is not None) or
                    (new_filter_value != new_value and new_filter_value is not None)):

                    return False

            filter_value = filter_attrs.get(name)
            if filter_value is not None:
                if isinstance(filter_value, list):  # A list of accepted values
                    if new_value not in filter_value:
                        return False

                elif new_value != filter_value:  # A single value
                    return False

        return True

    async def accepts_device(self, event: core_events.Event, old_attrs: Attributes, new_attrs: Attributes) -> bool:
        return self._accepts_attrs(
            self._filter_device_attr_names,
            self._filter_device_attrs,
            self._filter_device_attr_transitions,
            old_attrs,
            new_attrs
        )

    async def accepts_port_value(
        self,
        event: core_events.Event,
        value_pair: Tuple[NullablePortValue, NullablePortValue]
    ) -> bool:

        old_value, new_value = value_pair

        if self._filter_port_value_transition is not None:
            old_filter_value, new_filter_value = self._filter_port_value_transition
            if ((old_filter_value != old_value and old_filter_value is not None) or
                (new_filter_value != new_value and new_filter_value is not None)):

                return False

        if self._filter_port_value is not None:
            if isinstance(self._filter_port_value, list):  # A list of accepted values
                if new_value not in self._filter_port_value:
                    return False

            elif isinstance(self._filter_port_value, core_expressions.Expression):  # An expression
                if new_value != self._filter_port_value.eval():
                    return False

        return True

    async def accepts_port(
        self,
        event: core_events.Event,
        value_pair: Tuple[NullablePortValue, NullablePortValue],
        old_attrs: Attributes,
        new_attrs: Attributes
    ) -> bool:

        if not await self.accepts_port_value(event, value_pair):
            return False

        return self._accepts_attrs(
            self._filter_port_attr_names,
            self._filter_port_attrs,
            self._filter_port_attr_transitions,
            old_attrs,
            new_attrs
        )

    async def accepts_slave(self, event: core_events.Event, old_attrs: Attributes, new_attrs: Attributes) -> bool:
        return self._accepts_attrs(
            self._filter_slave_attr_names,
            self._filter_slave_attrs,
            self._filter_slave_attr_transitions,
            old_attrs,
            new_attrs
        )

    async def accepts(
        self,
        event: core_events.Event,
        value_pair: Tuple[NullablePortValue, NullablePortValue],
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> bool:

        if not self._filter_prepared:
            self._prepare_filter()

        if self._filter_event_types and event.get_type() not in self._filter_event_types:
            return False

        if (isinstance(event, core_events.DeviceEvent) and
            not await self.accepts_device(event, old_attrs, new_attrs)):

            return False

        elif (isinstance(event, core_events.PortEvent) and
              not await self.accepts_port(event, value_pair, old_attrs, new_attrs)):

            return False

        elif (isinstance(event, slaves_events.SlaveDeviceEvent) and
              not await self.accepts_slave(event, old_attrs, new_attrs)):

            return False

        return True

    async def handle_event(self, event: core_events.Event) -> None:
        (
            value_pair,
            old_attrs,
            new_attrs,
            changed_attrs,
            added_attrs,
            removed_attrs
        ) = await self._update_from_event(event)

        accepted = await self.accepts(
            event,
            value_pair,
            old_attrs,
            new_attrs,
            changed_attrs,
            added_attrs,
            removed_attrs
        )

        if not accepted:
            return

        self.debug('handling event %s', event)

        try:
            await self.on_event(event)

        except Exception as e:
            self.error('failed to handle event %s: %s', event, e, exc_info=True)

        try:
            if isinstance(event, core_events.ValueChange):
                old_value, new_value = value_pair
                await self.on_value_change(event, event.get_port(), old_value, new_value, new_attrs)

            elif isinstance(event, core_events.PortUpdate):
                await self.on_port_update(
                    event,
                    event.get_port(),
                    old_attrs,
                    new_attrs,
                    changed_attrs,
                    added_attrs,
                    removed_attrs
                )

            elif isinstance(event, core_events.PortAdd):
                await self.on_port_add(event, event.get_port(), new_attrs)

            elif isinstance(event, core_events.PortRemove):
                await self.on_port_remove(event, event.get_port(), new_attrs)

            elif isinstance(event, core_events.DeviceUpdate):
                await self.on_device_update(event, old_attrs, new_attrs, changed_attrs, added_attrs, removed_attrs)

            elif isinstance(event, core_events.FullUpdate):
                await self.on_full_update(event)

            elif isinstance(event, slaves_events.SlaveDeviceUpdate):
                await self.on_slave_device_update(
                    event,
                    event.get_slave(),
                    old_attrs,
                    new_attrs,
                    changed_attrs,
                    added_attrs,
                    removed_attrs
                )

            elif isinstance(event, slaves_events.SlaveDeviceAdd):
                await self.on_slave_device_add(event, event.get_slave(), new_attrs)

            elif isinstance(event, slaves_events.SlaveDeviceRemove):
                await self.on_slave_device_remove(event, event.get_slave(), new_attrs)

        except Exception as e:
            self.error('failed to handle event %s: %s', event, e, exc_info=True)

    async def on_event(self, event: core_events.Event) -> None:
        pass

    async def on_value_change(
        self,
        event: core_events.Event,
        port: core_ports.BasePort,
        old_value: NullablePortValue,
        new_value: NullablePortValue,
        attrs: Attributes
    ) -> None:

        pass

    async def on_port_update(
        self,
        event: core_events.Event,
        port: core_ports.BasePort,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> None:

        pass

    async def on_port_add(self, event: core_events.Event, port: core_ports.BasePort, attrs: Attributes) -> None:
        pass

    async def on_port_remove(self, event: core_events.Event, port: core_ports.BasePort, attrs: Attributes) -> None:
        pass

    async def on_device_update(
        self,
        event: core_events.Event,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> None:

        pass

    async def on_full_update(self, event: core_events.Event) -> None:
        pass

    async def on_slave_device_update(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave,
        old_attrs: Attributes,
        new_attrs: Attributes,
        changed_attrs: Dict[str, Tuple[Attribute, Attribute]],
        added_attrs: Attributes,
        removed_attrs: Attributes
    ) -> None:

        pass

    async def on_slave_device_add(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave,
        attrs: Attributes
    ) -> None:

        pass

    async def on_slave_device_remove(
        self,
        event: core_events.Event,
        slave: slaves_devices.Slave,
        attrs: Attributes
    ) -> None:

        pass




from typing import List

from qtoggleserver import persist
from qtoggleserver.core import api as core_api
from qtoggleserver.core.typing import GenericJSONDict


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
def get_panels(request: core_api.APIRequest) -> List[GenericJSONDict]:
    return persist.get_value('dashboard_panels', default=[])


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
def put_panels(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    # core_api.validate(panels, PANELS_SCHEMA)  TODO validate panels against schema

    persist.set_value('dashboard_panels', params)


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
def get_prefs(request: core_api.APIRequest) -> GenericJSONDict:
    return persist.get('ui_prefs', _id=request.username) or {}


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
def put_prefs(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    persist.replace('ui_prefs', _id=request.username, record=params)




import asyncio

from typing import Tuple

from qtoggleserver import version
from qtoggleserver.system import fwupdate


_SLEEP_TIME = 1
_DUMMY_VERSION = '1.0.2222-11-10'
_DUMMY_URL = f'http://firmware.qtoggle.io/qtoggleserver/{_DUMMY_VERSION}'
_DUMMY_DATE = '2222-11-10'


class DummyDriver(fwupdate.BaseDriver):
    def __init__(self) -> None:
        self._status = fwupdate.STATUS_IDLE

        fwupdate.BaseDriver.__init__(self)

    async def get_current_version(self) -> str:
        return version.VERSION

    async def get_latest(self) -> Tuple[str, str, str]:
        if self._status == fwupdate.STATUS_IDLE:
            await asyncio.sleep(_SLEEP_TIME)

        return _DUMMY_VERSION, _DUMMY_DATE, _DUMMY_URL

    async def get_status(self) -> str:
        return self._status

    async def update_to_version(self, version: str) -> None:
        asyncio.create_task(self._perform_update())

    async def update_to_url(self, url: str) -> None:
        asyncio.create_task(self._perform_update())

    async def _perform_update(self) -> None:
        self._status = fwupdate.STATUS_CHECKING
        await asyncio.sleep(_SLEEP_TIME)
        self._status = fwupdate.STATUS_DOWNLOADING
        await asyncio.sleep(_SLEEP_TIME)
        self._status = fwupdate.STATUS_EXTRACTING
        await asyncio.sleep(_SLEEP_TIME)
        self._status = fwupdate.STATUS_VALIDATING
        await asyncio.sleep(_SLEEP_TIME)
        self._status = fwupdate.STATUS_FLASHING
        await asyncio.sleep(_SLEEP_TIME)
        self._status = fwupdate.STATUS_RESTARTING
        await asyncio.sleep(_SLEEP_TIME * 2)
        self._status = fwupdate.STATUS_IDLE




import logging

from typing import Dict, List, Optional

from qtoggleserver import persist
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import GenericJSONDict, NullablePortValue, PortValue, PortValueChoices


logger = logging.getLogger(__name__)

_vport_args: Dict[str, GenericJSONDict] = {}


class VirtualPort(core_ports.Port):
    WRITABLE = True
    VIRTUAL = True

    def __init__(
        self,
        port_id: str,
        _type: str,
        _min: Optional[float],
        _max: Optional[float],
        integer: Optional[bool],
        step: Optional[float],
        choices: Optional[PortValueChoices]
    ) -> None:

        super().__init__(port_id)

        self._type: str = _type
        self._min: Optional[float] = _min
        self._max: Optional[float] = _max
        self._integer: Optional[bool] = integer
        self._step: Optional[float] = step
        self._choices: Optional[PortValueChoices] = choices

        self._value = self._virtual_value = self.adapt_value_type_sync(_type, integer, _min or 0)

    def map_id(self, new_id: str) -> None:
        raise core_ports.PortError('Virtual ports cannot be mapped')

    async def read_value(self) -> NullablePortValue:
        return self._virtual_value

    async def write_value(self, value: PortValue) -> None:
        self._virtual_value = value


def add(
    port_id: str,
    _type: str,
    _min: Optional[float],
    _max: Optional[float],
    integer: Optional[bool],
    step: Optional[float],
    choices: Optional[PortValueChoices]
) -> None:

    settings = {
        'id': port_id,
        'type': _type,
        'min': _min,
        'max': _max,
        'integer': integer,
        'step': step,
        'choices': choices
    }

    _vport_args[port_id] = settings

    logger.debug('saving virtual port settings for %s', port_id)
    persist.replace('vports', port_id, settings)


def remove(port_id: str) -> None:
    _vport_args.pop(port_id, None)
    logger.debug('removing virtual port settings for %s', port_id)
    persist.remove('vports', filt={'id': port_id})


def all_port_args() -> List[GenericJSONDict]:
    return [dict({'driver': VirtualPort, 'port_id': port_id}, **args)
            for port_id, args in _vport_args.items()]


def load() -> None:
    for entry in persist.query('vports'):
        _vport_args[entry['id']] = {
            '_type': entry.get('type') or core_ports.TYPE_NUMBER,
            '_min': entry.get('min'),
            '_max': entry.get('max'),
            'integer': entry.get('integer'),
            'step': entry.get('step'),
            'choices': entry.get('choices')
        }

        logger.debug('loaded virtual port settings for %s', entry['id'])


def reset() -> None:
    logger.debug('clearing virtual ports persisted data')
    persist.remove('vports')




from qtoggleserver.core.typing import PortValue as CorePortValue

from .functions import function, Function


@function('BITAND')
class BitAndFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        r = -1
        for e in self.eval_args():
            r &= int(e)

        return r


@function('BITOR')
class BitOrFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        r = 0
        for e in self.eval_args():
            r |= int(e)

        return r


@function('BITNOT')
class BitNotFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def eval(self) -> CorePortValue:
        return ~int(self.eval_args()[0])


@function('BITXOR')
class BitXOrFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        return int(eval_args[0]) ^ int(eval_args[1])


@function('SHL')
class SHLFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        return int(eval_args[0]) << int(eval_args[1])


@function('SHR')
class SHRFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def eval(self) -> CorePortValue:
        eval_args = self.eval_args()

        return int(eval_args[0]) >> int(eval_args[1])




from typing import Any

from qtoggleserver.core import api as core_api
from qtoggleserver.core.typing import GenericJSONDict

from .base import Event


# We can't use proper type annotations for ports in this module because that would create unsolvable circular imports.
# Therefore we use "Any" type annotation for BasePort instances.


class PortEvent(Event):
    def __init__(self, port: Any, timestamp: float = None) -> None:
        self._port = port

        super().__init__(timestamp)

    def __str__(self) -> str:
        return f'{self._type}({self._port.get_id()}) event'

    def get_port(self) -> Any:
        return self._port


class PortAdd(PortEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_VIEWONLY
    TYPE = 'port-add'

    async def get_params(self) -> GenericJSONDict:
        return await self.get_port().to_json()


class PortRemove(PortEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_VIEWONLY
    TYPE = 'port-remove'

    async def get_params(self) -> GenericJSONDict:
        return {'id': self.get_port().get_id()}


class PortUpdate(PortEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_VIEWONLY
    TYPE = 'port-update'

    async def get_params(self) -> GenericJSONDict:
        return await self.get_port().to_json()

    def is_duplicate(self, event: Event) -> bool:
        return isinstance(event, self.__class__) and event.get_port() == self.get_port()


class ValueChange(PortEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_VIEWONLY
    TYPE = 'value-change'

    async def get_params(self) -> GenericJSONDict:
        return {'id': self.get_port().get_id(), 'value': self.get_port().get_value()}




import datetime

from typing import Set

from qtoggleserver import system
from qtoggleserver.core.typing import PortValue as CorePortValue

from .functions import function, Function
from .exceptions import InvalidArgument, EvalSkipped


@function('YEAR')
class YearFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self) -> Set[str]:
        return {'time'}

    def eval(self) -> CorePortValue:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        return datetime.datetime.now().year


@function('MONTH')
class MonthFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self) -> Set[str]:
        return {'time'}

    def eval(self) -> CorePortValue:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        return datetime.datetime.now().month


@function('DAY')
class DayFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self) -> Set[str]:
        return {'time'}

    def eval(self) -> CorePortValue:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        return datetime.datetime.now().day


@function('DOW')
class DOWFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self) -> Set[str]:
        return {'time'}

    def eval(self) -> CorePortValue:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        return datetime.datetime.now().weekday()


@function('HOUR')
class HourFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self) -> Set[str]:
        return {'time'}

    def eval(self) -> CorePortValue:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        return datetime.datetime.now().hour


@function('MINUTE')
class MinuteFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self) -> Set[str]:
        return {'time'}

    def eval(self) -> CorePortValue:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        return datetime.datetime.now().minute


@function('SECOND')
class SecondFunction(Function):
    MIN_ARGS = MAX_ARGS = 0

    def get_deps(self) -> Set[str]:
        return {'time'}

    def eval(self) -> CorePortValue:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        return datetime.datetime.now().second


@function('HMSINTERVAL')
class HMSIntervalFunction(Function):
    MIN_ARGS = 4
    MAX_ARGS = 6

    def get_deps(self) -> Set[str]:
        return {'time'}

    def eval(self) -> CorePortValue:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.datetime.now()

        if len(self.args) >= 6:
            start_h = self.args[0].eval()
            start_m = self.args[1].eval()
            start_s = self.args[2].eval()
            stop_h = self.args[3].eval()
            stop_m = self.args[4].eval()
            stop_s = self.args[5].eval()

            if not (0 <= start_h <= 23):
                raise InvalidArgument(1, start_h)

            if not (0 <= start_m <= 59):
                raise InvalidArgument(2, start_m)

            if not (0 <= start_s <= 59):
                raise InvalidArgument(3, start_s)

            if not (0 <= stop_h <= 23):
                raise InvalidArgument(4, stop_h)

            if not (0 <= stop_m <= 59):
                raise InvalidArgument(5, stop_m)

            if not (0 <= stop_s <= 59):
                raise InvalidArgument(6, stop_s)

            start_time = datetime.time(start_h, start_m, start_s)
            stop_time = datetime.time(stop_h, stop_m, stop_s)

        else:  # Assuming 4
            start_h = self.args[0].eval()
            start_m = self.args[1].eval()
            stop_h = self.args[2].eval()
            stop_m = self.args[3].eval()

            if not (0 <= start_h <= 23):
                raise InvalidArgument(1, start_h)

            if not (0 <= start_m <= 59):
                raise InvalidArgument(2, start_m)

            if not (0 <= stop_h <= 23):
                raise InvalidArgument(3, stop_h)

            if not (0 <= stop_m <= 59):
                raise InvalidArgument(4, stop_m)

            start_time = datetime.time(start_h, start_m, 00)
            stop_time = datetime.time(stop_h, stop_m, 59)

        start_dt = datetime.datetime.combine(now.date(), start_time)
        stop_dt = datetime.datetime.combine(now.date(), stop_time)

        return start_dt <= now <= stop_dt




from typing import Optional

from qtoggleserver.core import ports
from qtoggleserver.utils import json as json_utils


class DummyGPIO(ports.Port):
    TYPE = ports.TYPE_BOOLEAN

    ADDITIONAL_ATTRDEFS = {
        'output': {
            'display_name': 'Is Output',
            'description': 'Controls the port direction.',
            'type': 'boolean',
            'modifiable': True
        }
    }

    def __init__(self, no: int, def_value: Optional[bool] = None, def_output: Optional[bool] = None) -> None:
        self._no: int = no

        self._def_value: Optional[bool] = def_value
        self._def_output: Optional[bool] = def_output

        self._dummy_value: bool = def_value or False
        self._dummy_output: bool = def_output if def_output is not None else False

        super().__init__(port_id=f'gpio{no}')

    async def handle_enable(self) -> None:
        if self._def_output is not None:
            await self.attr_set_output(self._def_output)

    async def read_value(self) -> bool:
        return self._dummy_value

    async def write_value(self, value: bool) -> None:
        self.debug('writing "%s"', json_utils.dumps(value))
        self._dummy_value = value

    async def attr_is_writable(self) -> bool:
        return self._dummy_output

    async def attr_set_output(self, output: bool) -> None:
        self._dummy_output = output

        if output:
            self.debug('setting output mode')

        else:
            self.debug('setting input mode')

        if output and self._def_value is not None:
            await self.write_value(self._def_value)

    async def attr_is_output(self) -> bool:
        return self._dummy_output




import abc
import asyncio
import functools
import inspect
import logging

from typing import Any, Callable, Dict, List, Optional, Type, Union

from qtoggleserver.core import ports as core_ports
from qtoggleserver.utils import logging as logging_utils

from .threadedrunner import ThreadedRunner


logger = logging.getLogger(__name__)


class Peripheral(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
    RUNNER_CLASS = ThreadedRunner
    RUNNER_QUEUE_SIZE = 8

    logger = logger

    def __init__(self, *, name: Optional[str] = None, **kwargs) -> None:
        logging_utils.LoggableMixin.__init__(self, name, self.logger)

        self._name: Optional[str] = name
        self._ports: List[core_ports.BasePort] = []
        self._enabled: bool = False
        self._online: bool = False
        self._runner: Optional[ThreadedRunner] = None
        self._port_update_task: Optional[asyncio.Task] = None

    def __str__(self) -> str:
        return f'peripheral {self.get_id()}'

    def get_id(self) -> str:
        return self.get_name() or f'{self.__class__.__name__}({hex(id(self))})'

    def get_name(self) -> Optional[str]:
        return self._name

    async def get_port_args(self) -> List[Dict[str, Any]]:
        port_args = self.make_port_args()
        # Compatibility shim for peripherals having non-awaitable make_port_args() method
        if inspect.isawaitable(port_args):
            port_args = await port_args

        # Transform port classes to dicts with drivers
        port_args = [
            {'driver': pa} if isinstance(pa, type) else pa
            for pa in port_args
        ]

        # Supply the peripheral argument
        for pa in port_args:
            pa.setdefault('peripheral', self)

        return port_args

    @abc.abstractmethod
    async def make_port_args(self) -> List[Union[Dict[str, Any], Type[core_ports.BasePort]]]:
        raise NotImplementedError()

    def set_ports(self, ports: List[core_ports.BasePort]) -> None:
        self._ports = ports

    def get_ports(self) -> List[core_ports.BasePort]:
        return list(self._ports)

    def is_enabled(self) -> bool:
        return self._enabled

    async def enable(self) -> None:
        if self._enabled:
            return

        self._enabled = True
        await self.handle_enable()
        self.debug('peripheral enabled')

    async def disable(self) -> None:
        if not self._enabled:
            return

        self._enabled = False
        await self.handle_disable()
        self.debug('peripheral disabled')

    async def check_disabled(self, exclude_port: Optional[core_ports.BasePort] = None) -> None:
        if not self._enabled:
            return

        for port in self._ports:
            if port.is_enabled() and port != exclude_port:
                break

        else:
            self.debug('all ports are disabled, disabling peripheral')
            await self.disable()

    def is_online(self) -> bool:
        return self._enabled and self._online

    def set_online(self, online: bool) -> None:
        if online and not self._online:
            self.debug('is online')
            self.handle_online()

        elif not online and self._online:
            self.debug('is offline')
            self.handle_offline()

        self._online = online

    def handle_offline(self) -> None:
        self.trigger_port_update_fire_and_forget()

    def handle_online(self) -> None:
        self.trigger_port_update_fire_and_forget()

    async def trigger_port_update(self) -> None:
        self._port_update_task = None
        for port in self._ports:
            if port.is_enabled():
                await port.trigger_update()

    def trigger_port_update_fire_and_forget(self) -> None:
        if self._port_update_task:
            return  # Already scheduled

        self._port_update_task = asyncio.create_task(self.trigger_port_update())

    def get_runner(self) -> ThreadedRunner:
        if self._runner is None:
            self._runner = self.make_runner()

        return self._runner

    def make_runner(self) -> ThreadedRunner:
        self.debug('starting threaded runner')
        runner = self.RUNNER_CLASS(queue_size=self.RUNNER_QUEUE_SIZE)
        runner.start()

        return runner

    async def run_threaded(self, func: Callable, *args, **kwargs) -> Any:
        future = asyncio.get_running_loop().create_future()

        def callback(result: Any, exception: Optional[Exception]) -> None:
            if exception:
                future.set_exception(exception)

            else:
                future.set_result(result)

        runner = self.get_runner()
        runner.schedule_func(functools.partial(func, *args, **kwargs), callback)

        return await future

    async def handle_enable(self) -> None:
        pass

    async def handle_disable(self) -> None:
        pass

    async def handle_init(self) -> None:
        pass

    async def handle_cleanup(self) -> None:
        if self._port_update_task:
            await self._port_update_task

        if self._runner:
            self.debug('stopping threaded runner')
            await self._runner.stop()
            self.debug('threaded runner stopped')



#!/usr/bin/env python

import asyncio
import code

from qtoggleserver import commands


def execute() -> None:
    import qtoggleserver  # noqa: F401; Required for locals()

    loop = asyncio.get_event_loop()

    loop.run_until_complete(commands.init())

    commands.logger.info('starting interactive shell')
    code.interact(local=locals())

    loop.run_until_complete(commands.cleanup())
    loop.run_until_complete(loop.shutdown_asyncgens())

    commands.logger.info('bye!')


if __name__ == '__main__':
    execute()




import typing as _typing


debug: bool = False

public_url: _typing.Optional[str] = None


logging: _typing.Dict[str, _typing.Any] = {
    'version': 1,
    'memory_logs_buffer_len': 10000,
    'formatters': {
        'default': {
            'format': '%(asctime)s: %(levelname)7s: [%(name)s] %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'default'
        }
    },
    'loggers': {
        # Double quotes are necessary to avoid HOCON key split
        '"qtoggleserver.core.sessions"': {'level': 'INFO'},
        '"qtoggleserver.persist"': {'level': 'INFO'},
        '"qtoggleserver.utils.cmd"': {'level': 'INFO'}
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['console']
    }
}


class core:

    class device_name:
        get_cmd: _typing.Optional[str] = None
        set_cmd: _typing.Optional[str] = None

    class passwords:
        set_cmd: _typing.Optional[str] = None

    tick_interval: int = 50
    event_queue_size: int = 256
    max_client_time_skew: int = 300
    listen_support: bool = True
    sequences_support: bool = True
    ssl_support: bool = True
    virtual_ports: int = 1024


class server:
    addr: str = '0.0.0.0'
    port: int = 8888
    compress_response: bool = True

    class https:
        cert_file: _typing.Optional[str] = None
        key_file: _typing.Optional[str] = None


class persist:
    driver: str = 'qtoggleserver.drivers.persist.JSONDriver'
    file_path: str = 'qtoggleserver-data.json'


class system:

    class date:
        set_cmd: _typing.Optional[str] = None
        set_format: _typing.Optional[str] = '%Y-%m-%dT%H:%M:%SZ'

    class timezone:
        get_cmd: _typing.Optional[str] = None
        set_cmd: _typing.Optional[str] = None

    class net:
        class wifi:
            get_cmd: _typing.Optional[str] = None
            set_cmd: _typing.Optional[str] = None

        class ip:
            get_cmd: _typing.Optional[str] = None
            set_cmd: _typing.Optional[str] = None

    class storage:
        path: _typing.Optional[str] = None

    class temperature:
        get_cmd: _typing.Optional[str] = None
        sensor_name: _typing.Optional[str] = None
        sensor_index: int = 0
        min: _typing.Optional[int] = 0
        max: _typing.Optional[int] = 100

    class battery:
        get_cmd: _typing.Optional[str] = None

    class fwupdate:
        driver: _typing.Optional[str] = None


class frontend:
    enabled: bool = True
    debug: bool = False
    static_url: str = None


class slaves:
    enabled: bool = True
    timeout: int = 10
    long_timeout: int = 60
    keepalive: int = 60
    retry_interval: int = 5
    retry_count: int = 3

    class discover:
        request_timeout: int = 5
        dhcp_timeout: int = 10
        dhcp_interface: str = None

        class ap:
            interface: str = None
            interface_cmd: str = None
            ssid: str = 'qToggleSetup'
            psk: str = None
            own_ip: str = '192.168.43.1'
            mask_len: int = 24
            start_ip: str = '192.168.43.50'
            stop_ip: str = '192.168.43.250'
            hostapd_binary: str = None
            hostapd_cli_binary: str = None
            dnsmasq_binary: str = None
            hostapd_log: str = '/tmp/hostapd.log'
            dnsmasq_log: str = '/tmp/dnsmasq.log'
            finish_timeout: int = 300


class webhooks:
    enabled: bool = False


class reverse:
    enabled: bool = False
    retry_interval: int = 5


event_handlers: _typing.List[_typing.Dict[str, _typing.Any]] = []

peripherals: _typing.List[_typing.Dict[str, _typing.Any]] = []

ports: _typing.List[_typing.Dict[str, _typing.Any]] = []

port_mappings: _typing.Dict[str, str] = {}




import time

from typing import List, Optional, Set, Tuple

from qtoggleserver.core.typing import PortValue as CorePortValue

from .exceptions import EvalSkipped
from .functions import function, Function


@function('DELAY')
class DelayFunction(Function):
    MIN_ARGS = MAX_ARGS = 2
    HISTORY_SIZE = 1024

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._queue: List[Tuple[int, CorePortValue]] = []
        self._last_value: Optional[CorePortValue] = None
        self._current_value: Optional[CorePortValue] = None

    def get_deps(self) -> Set[str]:
        return {'time_ms'}

    def eval(self) -> CorePortValue:
        time_ms = int(time.time() * 1000)

        value = self.args[0].eval()
        delay = self.args[1].eval()

        if self._current_value is None:
            self._current_value = value

        # Detect value transitions and build history
        if value != self._last_value:
            self._last_value = value

            # Drop elements from queue if history size reached
            while len(self._queue) >= self.HISTORY_SIZE:
                self._queue.pop(0)

            self._queue.append((time_ms, value))

        # Process history
        while self._queue and (time_ms - self._queue[0][0]) >= delay:
            self._current_value = self._queue.pop(0)[1]

        return self._current_value


@function('HELD')
class HeldFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._time_ms: Optional[int] = None
        self._last_value: Optional[CorePortValue] = None

    def get_deps(self) -> Set[str]:
        return {'time_ms'}

    def eval(self) -> CorePortValue:
        time_ms = int(time.time() * 1000)
        result = False

        value = self.args[0].eval()
        fixed_value = self.args[1].eval()
        duration = self.args[2].eval()

        if self._time_ms is None:  # Very first expression eval call
            self._time_ms = time_ms

        else:
            delta = time_ms - self._time_ms

            if self._last_value != value:
                self._time_ms = time_ms  # Reset held timer

            else:
                result = (delta >= duration) and (value == fixed_value)

        self._last_value = value

        return result


@function('DERIV')
class DerivFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_time: float = 0

    def get_deps(self) -> Set[str]:
        return {'time_ms'}

    def eval(self) -> CorePortValue:
        value = self.args[0].eval()
        sampling_interval = self.args[1].eval() / 1000
        result = 0
        now = time.time()

        if self._last_value is not None:
            delta = now - self._last_time
            if delta < sampling_interval:
                raise EvalSkipped()

            result = (value - self._last_value) / delta

        self._last_value = value
        self._last_time = now

        return result


@function('INTEG')
class IntegFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_time: float = 0

    def get_deps(self) -> Set[str]:
        return {'time_ms'}

    def eval(self) -> CorePortValue:
        value = self.args[0].eval()
        accumulator = self.args[1].eval()
        sampling_interval = self.args[2].eval() / 1000
        result = accumulator
        now = time.time()

        if self._last_value is not None:
            delta = now - self._last_time
            if delta < sampling_interval:
                raise EvalSkipped()

            result += (value + self._last_value) * delta / 2

        self._last_value = value
        self._last_time = now

        return result


@function('FMAVG')
class FMAvgFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    QUEUE_SIZE = 1024

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._queue: List[CorePortValue] = []
        self._last_time: float = 0

    def get_deps(self) -> Set[str]:
        return {'time_ms'}

    def eval(self) -> CorePortValue:
        value = self.args[0].eval()
        width = min(self.args[1].eval(), self.QUEUE_SIZE)
        sampling_interval = self.args[2].eval() / 1000
        now = time.time()

        if self._last_time > 0:
            delta = now - self._last_time
            if delta < sampling_interval:
                raise EvalSkipped()

        # Make place for the new element
        while len(self._queue) > width:
            self._queue.pop(0)

        self._queue.append(value)
        self._last_time = now

        return sum(self._queue) / len(self._queue)


@function('FMEDIAN')
class FMedianFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    QUEUE_SIZE = 1024

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._queue: List[CorePortValue] = []
        self._last_time: float = 0

    def get_deps(self) -> Set[str]:
        return {'time_ms'}

    def eval(self) -> CorePortValue:
        value = self.args[0].eval()
        width = min(self.args[1].eval(), self.QUEUE_SIZE)
        sampling_interval = self.args[2].eval() / 1000
        now = time.time()

        if self._last_time > 0:
            delta = now - self._last_time
            if delta < sampling_interval:
                raise EvalSkipped()

        # Make place for the new element
        while len(self._queue) > width:
            self._queue.pop(0)

        self._queue.append(value)
        self._last_time = now

        sorted_queue = list(sorted(self._queue))

        return sorted_queue[len(sorted_queue) // 2]




from __future__ import annotations

import abc
import asyncio
import functools
import logging
import re
import subprocess
import time

from typing import Any, Callable, Dict, List, Optional, Tuple

from bluepy import btle

from qtoggleserver.core import ports as core_ports
from qtoggleserver.peripherals import Peripheral, RunnerBusy
from qtoggleserver.utils import logging as logging_utils

from . import polled


logger = logging.getLogger(__name__)

DEFAULT_ADAPTER_NAME = 'hci0'


class BLEException(Exception):
    pass


class BLEBusy(BLEException):
    def __init__(self, message: str = 'busy') -> None:
        super().__init__(message)


class BLETimeout(BLEException):
    def __init__(self, message: str = 'timeout') -> None:
        super().__init__(message)


class _BluepyTimeoutError(btle.BTLEException):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class _BluepyPeripheral(btle.Peripheral):
    def __init__(self, timeout: Optional[int] = None, *args, **kwargs) -> None:
        self._timeout: Optional[int] = timeout
        super().__init__(*args, **kwargs)

    def _getResp(self, wantType: Any, timeout: int = None) -> Any:
        # Override this method to be able to inject a default timeout
        # We also need to raise a timeout exception in case the response is None

        timeout = timeout or self._timeout
        response = super()._getResp(wantType, timeout)
        if response is None:
            raise _BluepyTimeoutError('Timeout waiting for a response from peripheral')

        return response

    def _stopHelper(self) -> None:
        # Override this method to close the helper's stdout and stdin streams

        helper = self._helper
        super()._stopHelper()
        self._helper = helper
        if helper:
            helper.stdin.close()
            helper.stdout.close()


class BLEAdapter(logging_utils.LoggableMixin):
    RUNNER_CLASS = Peripheral.RUNNER_CLASS

    _adapters_by_name: Dict[str, BLEAdapter] = {}

    @classmethod
    def get(cls, name: str) -> BLEAdapter:
        if name not in cls._adapters_by_name:
            logger.debug('initializing adapter %s', name)
            adapter = cls(name)
            cls._adapters_by_name[name] = adapter

        return cls._adapters_by_name[name]

    def __init__(self, name: str) -> None:
        logging_utils.LoggableMixin.__init__(self, name, logger)

        self._name: str = name
        try:
            self._address: str = self._find_address(name)

        except Exception as e:
            self.error('could not initialize adapter: %s', e)
            raise

        self.debug('found adapter address %s', self._address)

        self._runner: Optional[BLEPeripheral.RUNNER_CLASS] = None
        self._peripherals: List[BLEPeripheral] = []

    def __str__(self) -> str:
        return self._name

    def add_peripheral(self, peripheral: BLEPeripheral) -> None:
        self._peripherals.append(peripheral)

    def get_runner(self) -> BLEPeripheral.RUNNER_CLASS:
        if self._runner is None:
            self._runner = self.make_runner()

        return self._runner

    def make_runner(self) -> BLEPeripheral.RUNNER_CLASS:
        self.debug('starting threaded runner')
        runner = self.RUNNER_CLASS(queue_size=Peripheral.RUNNER_QUEUE_SIZE)
        runner.start()

        return runner

    @staticmethod
    def pretty_data(data: bytes) -> str:
        return ' '.join(map(lambda c: f'{c:02X}', data))

    @staticmethod
    def _find_address(name: str) -> str:
        try:
            output = subprocess.check_output(['hcitool', '-i', name, 'dev'], stderr=subprocess.STDOUT).decode()

        except subprocess.CalledProcessError as e:
            output = e.output.decode()
            if output.count('No such device'):
                raise Exception(f'Adapter not found: {name}') from e

            else:
                raise Exception(output.strip()) from e

        found = re.findall(name + r'\s([0-9a-f:]{17})', output, re.IGNORECASE)
        if not found:
            raise Exception(f'Adapter not found: {name}')

        return found[0]


class BLEPeripheral(polled.PolledPeripheral, metaclass=abc.ABCMeta):
    CMD_TIMEOUT = 10
    RETRY_COUNT = 2
    RETRY_DELAY = 2

    logger = logger

    def __init__(self, *, address: str, adapter_name: str = DEFAULT_ADAPTER_NAME, **kwargs) -> None:
        self._address = address
        self._adapter: BLEAdapter = BLEAdapter.get(adapter_name)
        self._adapter.add_peripheral(self)

        super().__init__(**kwargs)

    def __str__(self) -> str:
        return f'{self._adapter}/{self._name}'

    def make_runner(self) -> BLEPeripheral.RUNNER_CLASS:
        # Instead of creating a runner for each peripheral instance, we use adapter's runner, thus having one runner per
        # adapter instance.

        return self._adapter.get_runner()

    async def read(self, handle: int, retry_count: Optional[int] = None) -> Tuple[Optional[bytes], Optional[bytes]]:
        if retry_count is None:
            retry_count = self.RETRY_COUNT

        return await self._run_cmd_async(
            'read',
            self._address,
            handle,
            notify_handle=None,
            data=None,
            timeout=self.CMD_TIMEOUT,
            retry_count=retry_count
        )

    async def write(
        self,
        handle: int,
        data: bytes,
        retry_count: Optional[int] = None
    ) -> Tuple[Optional[bytes], Optional[bytes]]:

        if retry_count is None:
            retry_count = self.RETRY_COUNT

        return await self._run_cmd_async(
            'write',
            self._address,
            handle,
            notify_handle=None,
            data=data,
            timeout=self.CMD_TIMEOUT,
            retry_count=retry_count
        )

    async def write_notify(
        self,
        handle: int,
        notify_handle: int,
        data: bytes,
        timeout: Optional[int] = None,
        retry_count: int = None
    ) -> Tuple[Optional[bytes], Optional[bytes]]:

        if retry_count is None:
            retry_count = self.RETRY_COUNT

        if timeout is None:
            timeout = self.CMD_TIMEOUT

        return await self._run_cmd_async(
            'write',
            self._address,
            handle,
            notify_handle=notify_handle,
            data=data,
            timeout=timeout,
            retry_count=retry_count
        )

    def _run_cmd(
        self,
        cmd: str,
        address: str,
        handle: int,
        notify_handle: Optional[int],
        data: Optional[bytes],
        timeout: Optional[int]
    ) -> Tuple[Optional[bytes], Optional[bytes]]:

        start_time = time.time()
        response = None
        notification_data = None

        self.debug('connecting to %s', address)

        bluepy_peripheral = _BluepyPeripheral(timeout=timeout)
        try:
            bluepy_peripheral.connect(address)

        except _BluepyTimeoutError:
            bluepy_peripheral._stopHelper()
            raise BLETimeout(f'Timeout connecting to {address}') from None

        if notify_handle:
            # Create a temporary class to deal with BTLE delegation

            peripheral = self

            class Delegate(btle.DefaultDelegate):

                def handleNotification(self, h: int, d: bytes) -> None:
                    nonlocal notification_data

                    if not notify_handle or h != notify_handle:
                        peripheral.warning('got unexpected notification on %04X', h)
                        return

                    peripheral.debug('got notification on %04X: %s', h, BLEPeripheral.pretty_data(d))
                    notification_data = d

            bluepy_peripheral.withDelegate(Delegate())

        if cmd == 'write':
            self.debug('writing at %04X: %s', handle, self.pretty_data(data))
            bluepy_peripheral.writeCharacteristic(handle, data, withResponse=True)

        else:  # Assuming read
            self.debug('reading from %04X', handle)
            response = bluepy_peripheral.readCharacteristic(handle)
            self.debug('got response: %s', self.pretty_data(response))

        if notify_handle and notification_data is None:
            self.debug('waiting for notification on %04X', notify_handle)

            while not notification_data and self.get_runner().is_running():
                if time.time() - start_time > timeout:
                    bluepy_peripheral._stopHelper()
                    raise BLETimeout(f'Timeout waiting for notification on {notify_handle:04X} from {address}')

                try:
                    bluepy_peripheral.waitForNotifications(0.1)

                except _BluepyTimeoutError:
                    continue

        self.debug('%s command done', cmd)

        return response, notification_data

    async def _run_cmd_async(
        self,
        cmd: str,
        address: str,
        handle: int,
        notify_handle: Optional[int],
        data: Optional[bytes],
        timeout: Optional[int],
        retry_count: int
    ) -> Tuple[Optional[bytes], Optional[bytes]]:

        retry = 1
        while True:
            try:
                response, notification_data = await self.run_threaded(
                    self._run_cmd,
                    cmd,
                    address,
                    handle,
                    notify_handle,
                    data,
                    timeout
                )

            except Exception as e:
                self.error('command execution failed: %s', e)

                if retry <= retry_count:
                    await asyncio.sleep(self.RETRY_DELAY)
                    self.warning('retry %s/%s', retry, retry_count)
                    retry += 1
                    continue

                self.set_online(False)

                if isinstance(e, RunnerBusy):
                    raise BLEBusy('Too many pending commands') from e

                raise

            else:
                self.set_online(True)

                return response, notification_data

    @staticmethod
    def pretty_data(data: bytes) -> str:
        return BLEAdapter.pretty_data(data)


class BLEPort(polled.PolledPort, metaclass=abc.ABCMeta):
    READ_INTERVAL_MAX = 1440
    READ_INTERVAL_STEP = 1
    READ_INTERVAL_MULTIPLIER = 60


def port_exceptions(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        # Transform BLE exceptions into port exceptions, where applicable
        try:
            return func(*args, **kwargs)

        except BLETimeout as e:
            raise core_ports.PortTimeout() from e

        except BLEException as e:
            raise core_ports.PortError(str(e)) from e

    return wrapper




from __future__ import annotations

import abc
import asyncio
import queue
import threading

from typing import Callable, Optional


class RunnerBusy(Exception):
    pass


class ThreadedRunner(threading.Thread, metaclass=abc.ABCMeta):
    QUEUE_TIMEOUT = 1

    def __init__(self, queue_size: Optional[int] = None) -> None:
        self._running: bool = False
        self._loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        self._queue: queue.Queue = queue.Queue(queue_size or 0)
        self._queue_size: int = queue_size
        self._stopped_future: asyncio.Future = self._loop.create_future()

        super().__init__()

    def run(self) -> None:
        while self._running:
            try:
                func, callback = self._queue.get(timeout=self.QUEUE_TIMEOUT)

            except queue.Empty:
                continue

            try:
                result = func()

            except Exception as e:
                self._loop.call_soon_threadsafe(callback, None, e)

            else:
                self._loop.call_soon_threadsafe(callback, result, None)

        self._loop.call_soon_threadsafe(self._stopped_future.set_result, None)

    def schedule_func(self, func: Callable, callback: Callable) -> None:
        try:
            self._queue.put_nowait((func, callback))

        except queue.Full:
            raise RunnerBusy() from None

    def is_running(self) -> bool:
        return self._running

    def start(self) -> None:
        self._running = True
        super().start()

    async def stop(self) -> None:
        self._running = False

        await self._stopped_future




import asyncio
import logging
import psutil
import subprocess
import tempfile
import time

from typing import Dict, List, Optional, TextIO, Union

from .exceptions import APException


BINARY = 'dnsmasq'

DNSMASQ_CONF_TEMPLATE = (
    'interface={interface}\n'
    'dhcp-range={start_ip},{stop_ip},24h\n'
    'dhcp-leasefile={leases_file}\n'
    'no-ping\n'
)

STOP_TIMEOUT = 2


logger = logging.getLogger(__name__)


class DNSMasqException(APException):
    pass


class DNSMasq:
    def __init__(
        self,
        interface: str,
        own_ip: str,
        mask_len: int,
        start_ip: str,
        stop_ip: str,
        dnsmasq_binary: Optional[str] = None,
        dnsmasq_log: Optional[str] = None
    ) -> None:

        self._interface: str = interface
        self._own_ip: str = own_ip
        self._mask_len: int = mask_len
        self._start_ip: str = start_ip
        self._stop_ip: str = stop_ip
        self._binary: Optional[str] = dnsmasq_binary
        self._log: Optional[str] = dnsmasq_log

        self._conf_file: Optional[TextIO] = None
        self._log_file: Optional[TextIO] = None
        self._leases_file: Optional[TextIO] = None
        self._process: Optional[subprocess.Popen] = None

        self._leases: List[Dict[str, Union[str, int]]] = []

    def is_alive(self) -> bool:
        return (self._process is not None) and (self._process.poll() is None)

    def is_running(self) -> bool:
        return self._process is not None

    def start(self) -> None:
        logger.debug(
            'starting dnsmasq with IP range %s - %s and own IP %s/%d',
            self._start_ip,
            self._stop_ip,
            self._own_ip,
            self._mask_len
        )

        binary = self._binary or self._find_binary()
        if not binary:
            raise DNSMasqException('Could not find %s binary', BINARY)

        self.ensure_own_ip()

        self._leases_file = tempfile.NamedTemporaryFile(mode='w+t')

        conf = DNSMASQ_CONF_TEMPLATE.format(
            start_ip=self._start_ip,
            stop_ip=self._stop_ip,
            interface=self._interface,
            leases_file=self._leases_file.name
        )

        self._log_file = open(self._log, 'wt')
        self._conf_file = tempfile.NamedTemporaryFile(mode='wt')
        self._conf_file.write(conf)
        self._conf_file.flush()

        self._process = subprocess.Popen(
            [binary, '-d', '-C', self._conf_file.name],
            stdout=self._log_file,
            stderr=subprocess.STDOUT
        )

    async def stop(self) -> None:
        logger.debug('stopping dnsmasq')

        if self._process:
            self._process.terminate()

            begin_time = time.time()
            while self._process.poll() is None:
                await asyncio.sleep(0.1)
                if time.time() - begin_time > STOP_TIMEOUT:
                    break

            # If process could not be stopped in time, kill it
            if self._process.poll() is None:
                logger.error('failed to stop hostapd within %d seconds, killing it', STOP_TIMEOUT)
                self._process.kill()
                await asyncio.sleep(1)
                self._process.poll()  # We want no zombies

            self._process = None

        if self._conf_file:
            self._conf_file.close()
            self._conf_file = None

        if self._log_file:
            self._log_file.close()
            self._log_file = None

        if self._leases_file:
            # Read leases file just before removing it
            self._leases = self._read_leases_file()

            self._leases_file.close()
            self._leases_file = None

    def ensure_own_ip(self) -> None:
        try:
            subprocess.check_call(
                ['ip', 'addr', 'flush', 'dev', self._interface],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )

        except subprocess.CalledProcessError:
            raise DNSMasqException('Could not clear current own IP address')

        try:
            subprocess.check_call(
                ['ip', 'addr', 'add', f'{self._own_ip}/{self._mask_len}', 'dev', self._interface],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )

        except subprocess.CalledProcessError:
            raise DNSMasqException('Could not set own IP address')

    def _find_binary(self) -> Optional[str]:
        try:
            return subprocess.check_output(['which', BINARY], stderr=subprocess.DEVNULL).decode().strip()

        except subprocess.CalledProcessError:
            return None

    def _read_leases_file(self) -> List[Dict[str, Union[str, int]]]:
        self._leases_file.seek(0)
        lines = self._leases_file.readlines()

        leases = []
        for line in lines:
            line = line.strip()
            if not line:
                continue

            parts = line.split()
            if len(parts) < 4:
                continue

            timestamp, mac_address, ip_address, hostname = parts[:4]
            leases.append({
                'timestamp': int(timestamp),
                'mac_address': mac_address,
                'ip_address': ip_address,
                'hostname': hostname
            })

        return leases

    def get_leases(self) -> List[Dict[str, Union[str, int]]]:
        if self._leases_file:
            self._leases = self._read_leases_file()

        return self._leases

    @staticmethod
    def is_already_running() -> bool:
        return BINARY in (p.name() for p in psutil.process_iter())




import copy
import logging
import os

from typing import Any, Dict, Iterable, List, Optional

from qtoggleserver.conf import settings
from qtoggleserver.persist import BaseDriver, Id, Record
from qtoggleserver.utils import json as json_utils


DEFAULT_FILE_PATH = 'qtoggleserver-data.json'

logger = logging.getLogger(__name__)

Collection = Dict[int, Record]
IndexedData = Dict[str, Collection]
UnindexedData = Dict[str, List[Record]]


class JSONPersistError(Exception):
    pass


class DuplicateRecordId(JSONPersistError):
    pass


class JSONDriver(BaseDriver):
    def __init__(
        self,
        file_path: str = DEFAULT_FILE_PATH,
        pretty_format: Optional[bool] = None,
        use_backup: bool = True,
        **kwargs
    ) -> None:

        # Pretty formatting follows general debug flag, by default
        if pretty_format is None:
            pretty_format = settings.debug

        self._file_path: str = file_path
        self._pretty_format: bool = pretty_format
        self._use_backup: bool = use_backup

        self._data: IndexedData = self._index(self._load())

    def query(
        self,
        collection: str,
        fields: Optional[List[str]],
        filt: Dict[str, Any],
        limit: Optional[int]
    ) -> Iterable[Record]:

        coll = self._data.get(collection, {})
        records = []

        if 'id' in filt:  # Look for specific record id
            filt = dict(filt)
            _id = filt.pop('id')
            record = coll.get(_id)

            # Apply filter criteria
            if record is not None and self._filter_matches(record, filt):
                records.append(record)

        else:
            for record in coll.values():
                # Apply filter criteria
                if self._filter_matches(record, filt):
                    records.append(record)

        # Apply limit
        if limit is not None:
            records = records[:limit]

        # Apply projection
        if fields is not None:
            fields = set(fields)
            projected_records = []
            for record in records:
                projected_record = {k: v for k, v in record.items() if k in fields}
                projected_records.append(projected_record)

            records = projected_records

        return copy.deepcopy(records)

    def insert(self, collection: str, record: Record) -> Id:
        coll = self._data.setdefault(collection, {})

        _id = record.get('id')
        if _id is None:
            _id = self._find_next_id(coll)
            record = dict(record, id=_id)

        elif _id in coll:
            raise DuplicateRecordId(_id)

        coll[_id] = record

        self._save(self._unindex(self._data))

        return _id

    def update(self, collection: str, record_part: Record, filt: Dict[str, Any]) -> int:
        coll = self._data.setdefault(collection, {})
        modified_count = 0

        if 'id' in filt:
            filt = dict(filt)
            _id = filt.pop('id')

            record = coll.get(_id)
            if record is not None:
                record.update(record_part)
                modified_count = 1

        else:  # No id in filt
            for record in coll.values():
                # Apply filter criteria
                if not self._filter_matches(record, filt):
                    continue

                # Actually update the record
                record.update(record_part)
                modified_count += 1

        self._save(self._unindex(self._data))

        return modified_count

    def replace(self, collection: str, _id: Id, record: Record, upsert: bool) -> bool:
        coll = self._data.setdefault(collection, {})

        record_existed = coll.get(_id) is not None
        if record_existed is None and not upsert:
            return False  # No record found, no replacing

        # Remove old id if a new id is supplied
        new_id = record.get('id', None)
        if (new_id is not None) and (new_id != _id) and record_existed:
            coll.pop(_id)

        if new_id is None:
            new_id = _id
            record['id'] = _id

        coll[new_id] = record

        self._save(self._unindex(self._data))

        return record_existed

    def remove(self, collection: str, filt: Dict[str, Any]) -> int:
        coll = self._data.setdefault(collection, {})
        removed_count = 0

        if 'id' in filt:
            filt = dict(filt)
            _id = filt.pop('id')

            record = coll.get(_id)
            if (record is not None) and self._filter_matches(record, filt):
                coll.pop(_id)
                removed_count = 1

        else:  # No id in filt
            for _id, record in list(coll.items()):
                # Apply filter criteria
                if not self._filter_matches(record, filt):
                    continue

                # Actually remove the record
                coll.pop(_id)
                removed_count += 1

        self._save(self._unindex(self._data))

        return removed_count

    def cleanup(self) -> None:
        pass

    @staticmethod
    def _filter_matches(record: Record, filt: Dict[str, Any]) -> bool:
        for key, value in filt.items():
            try:
                if record[key] != value:
                    return False

            except KeyError:
                return False

        return True

    @staticmethod
    def _find_next_id(coll: Collection) -> Id:
        int_ids = [int(_id) for _id in coll.keys() if _id]
        int_ids.append(0)

        return str(max(int_ids) + 1)

    def _get_backup_file_path(self) -> str:
        path, ext = os.path.splitext(self._file_path)
        return f'{path}_backup{ext}'

    def _load(self) -> UnindexedData:
        logger.debug('loading from %s', self._file_path)

        try:
            # If the file is accessible but empty, consider data loaded and return empty dictionary
            if os.stat(self._file_path).st_size == 0:
                logger.debug('file %s is empty', self._file_path)
                return {}

        except FileNotFoundError:
            # If the file does not exist, consider data loaded and return empty dictionary
            logger.debug('file %s does not exist', self._file_path)
            return {}

        try:
            with open(self._file_path, 'rb') as f:
                data = f.read()
                return json_utils.loads(data, allow_extended_types=True)

        except Exception as e:
            if not self._use_backup:
                raise

            # Upon failure, if using a backup, simply log the error and attempt to load from backup file
            logger.error('failed to load from %s: %s', self._file_path, e, exc_info=True)

            backup_file_path = self._get_backup_file_path()
            logger.warning('loading from backup %s', backup_file_path)

            with open(backup_file_path, 'rb') as f:
                return json_utils.loads(f.read(), allow_extended_types=True)

    def _save(self, data: UnindexedData) -> None:
        if self._use_backup and os.path.exists(self._file_path):
            backup_file_path = self._get_backup_file_path()
            logger.debug('backing up %s to %s', self._file_path, backup_file_path)
            os.rename(self._file_path, backup_file_path)

        logger.debug('saving to %s', self._file_path)

        with open(self._file_path, 'wb') as f:
            data = json_utils.dumps(data, allow_extended_types=True, indent=4 if self._pretty_format else None)
            f.write(data.encode())

    @staticmethod
    def _index(data: UnindexedData) -> IndexedData:
        indexed_data = {}
        for coll, records in data.items():
            indexed_data[coll] = {r.get('id', ''): r for r in records}

        return indexed_data

    @staticmethod
    def _unindex(data: IndexedData) -> UnindexedData:
        unindexed_data = {}
        for coll, records in data.items():
            unindexed_data[coll] = list(records.values())

        return unindexed_data




from typing import List

from qtoggleserver.core import api as core_api
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.typing import GenericJSONDict
from qtoggleserver.slaves import discover as slaves_discover
from qtoggleserver.slaves import exceptions as slaves_exceptions

from .. import schema as api_schema


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_discovered(request: core_api.APIRequest, timeout: int) -> List[GenericJSONDict]:
    if timeout is None:
        raise core_api.APIError(400, 'missing-field', field='timeout')

    discovered_devices = slaves_discover.get_discovered_devices()
    if discovered_devices is None:
        await slaves_discover.discover(timeout)

    return [d.to_json() for d in slaves_discover.get_discovered_devices().values()]


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def delete_discovered(request: core_api.APIRequest) -> None:
    await slaves_discover.finish()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_discovered_device(request: core_api.APIRequest, name: str, params: GenericJSONDict) -> GenericJSONDict:
    core_api_schema.validate(params, api_schema.PATCH_DISCOVERED_DEVICE)

    discovered_devices = slaves_discover.get_discovered_devices() or {}
    discovered_device = discovered_devices.get(name)
    if not discovered_device:
        raise core_api.APIError(404, 'no-such-device')

    attrs = params['attrs']
    try:
        discovered_device = await slaves_discover.configure(discovered_device, attrs)

    except Exception as e:
        raise slaves_exceptions.adapt_api_error(e) from e

    return discovered_device.to_json()




import datetime
import hashlib
import logging

from typing import Any, Optional

from tornado import httpclient

from qtoggleserver.conf import settings
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils import logging as logging_utils


logger = logging.getLogger(__name__)


class APClient(logging_utils.LoggableMixin):
    def __init__(
        self,
        mac_address: str,
        ip_address: str,
        hostname: str,
        moment: datetime
    ) -> None:

        self.mac_address: str = mac_address
        self.ip_address: str = ip_address
        self.hostname: str = hostname
        self.moment: datetime = moment

        super().__init__(self.mac_address, logger)

    def __str__(self) -> str:
        return f'APClient {self.mac_address} at {self.ip_address}'

    async def request(
        self, method: str,
        path: str,
        body: Any = None,
        admin_password: Optional[str] = None,
        no_log: bool = False
    ) -> Any:

        http_client = httpclient.AsyncHTTPClient()
        if admin_password:
            password_hash = hashlib.sha256(admin_password.encode()).hexdigest()

        else:
            password_hash = core_api_auth.EMPTY_PASSWORD_HASH

        headers = {
            'Content-Type': json_utils.JSON_CONTENT_TYPE,
            'Authorization': core_api_auth.make_auth_header(
                core_api_auth.ORIGIN_CONSUMER,
                username='admin',
                password_hash=password_hash
            )
        }

        # TODO: this only tries standard port 80; ideally it should try connecting on a list of known ports

        timeout = settings.slaves.discover.request_timeout
        url = f'http://{self.ip_address}:80{path}'

        body_str = None
        if body is not None:
            body_str = json_utils.dumps(body)

        request = httpclient.HTTPRequest(
            url=url,
            method=method,
            headers=headers,
            body=body_str,
            connect_timeout=timeout,
            request_timeout=timeout
        )

        if not no_log:
            self.debug('requesting %s %s', method, path)

        try:
            response = await http_client.fetch(request, raise_error=True)

        except Exception as e:
            if not no_log:
                self.error('request %s %s failed: %s', method, path, e, exc_info=True)
            raise

        if response.body:
            return json_utils.loads(response.body)




import abc

from typing import Callable, List, Optional, Set

from qtoggleserver.core.typing import PortValue as CorePortValue

from . import exceptions
from . import parse
from .base import Expression


FUNCTIONS = {}


def function(name: str) -> Callable:
    def decorator(func_class: type) -> type:
        func_class.NAME = name
        FUNCTIONS[name] = func_class

        return func_class

    return decorator


class Function(Expression, metaclass=abc.ABCMeta):
    NAME = None
    MIN_ARGS = None
    MAX_ARGS = None

    def __init__(self, args: List[Expression]) -> None:
        self.args: List[Expression] = args
        self._deps: Optional[Set[str]] = None

    def __str__(self) -> str:
        s = getattr(self, '_str', None)
        if s is None:
            args_str = ', '.join(str(e) for e in self.args)
            self._str = s = f'{self.NAME}({args_str})'

        return s

    def get_deps(self) -> Set[str]:
        if self._deps is None:
            self._deps = set()

            for arg in self.args:
                self._deps |= arg.get_deps()

        return self._deps

    def eval_args(self) -> List[CorePortValue]:
        return [a.eval() for a in self.args]

    @staticmethod
    def parse(self_port_id: Optional[str], sexpression: str, pos: int) -> Expression:
        while sexpression and sexpression[0].isspace():
            sexpression = sexpression[1:]
            pos += 1

        while sexpression and sexpression[-1].isspace():
            sexpression = sexpression[:-1]

        p_start = None
        p_end = None
        p_last_comma = None
        level = 0
        sargs = []
        for i, c in enumerate(sexpression):
            if c == '(':
                if p_start is None:
                    p_start = i

                elif level == 0:
                    raise exceptions.UnexpectedCharacter(c, pos + i)

                level += 1

            elif c == ')':
                if level == 0:
                    raise exceptions.UnbalancedParentheses(pos + i)

                elif level == 1:
                    if p_end is None:
                        p_end = i

                    else:
                        raise exceptions.UnbalancedParentheses(pos + i)

                level -= 1

            elif (c == ',') and (level == 1):
                sarg = sexpression[(p_last_comma or p_start) + 1: i]
                spos = (p_last_comma or p_start) + 1
                if not sarg.strip():
                    raise exceptions.UnexpectedCharacter(c, pos + spos + len(sarg))

                sargs.append((sarg, spos))
                p_last_comma = i

            elif (p_start is not None) and (level == 0) and not c.isspace():
                raise exceptions.UnexpectedCharacter(c, pos + i)

        if (p_start is None) or (p_end is None) or (p_start > p_end) or (level != 0):
            raise exceptions.UnexpectedEnd()

        if p_end - p_start > 1:
            sarg = sexpression[(p_last_comma or p_start) + 1: p_end]
            spos = (p_last_comma or p_start) + 1
            if not sarg.strip():
                raise exceptions.UnexpectedCharacter(')', pos + spos + len(sarg))

            sargs.append((sarg, spos))

        func_name = sexpression[:p_start].strip()
        func_class = FUNCTIONS.get(func_name)
        if func_class is None:
            raise exceptions.UnknownFunction(func_name, pos)

        if func_class.MIN_ARGS is not None and len(sargs) < func_class.MIN_ARGS:
            raise exceptions.InvalidNumberOfArguments(func_name, pos)

        if func_class.MAX_ARGS is not None and len(sargs) > func_class.MAX_ARGS:
            raise exceptions.InvalidNumberOfArguments(func_name, pos)

        args = [parse(self_port_id, sarg, pos + spos) for (sarg, spos) in sargs]

        return func_class(args)


# These imports are here just because we need all modules to be imported, so that @function decorator registers all
# defined functions.

from .arithmetic import function as _
from .logic import function as _
from .bitwise import function as _
from .comparison import function as _
from .sign import function as _
from .aggregation import function as _
from .rounding import function as _
from .time import function as _
from .date import function as _
from .timeprocessing import function as _
from .various import function as _




import types

from collections import OrderedDict
from typing import Any, Dict

import pyhocon


_config_factory = pyhocon.ConfigFactory()


def obj_to_dict(obj: Any) -> Dict[str, Any]:
    d = {}
    for k, v in obj.__dict__.items():
        if k.startswith('__') or isinstance(v, types.ModuleType):
            continue

        if isinstance(v, type):
            v = obj_to_dict(v)

        d[k] = v

    return d


def update_obj_from_dict(obj: Any, d: OrderedDict) -> None:
    for k, v in d.items():
        ov = getattr(obj, k, None)
        if isinstance(ov, type):
            update_obj_from_dict(ov, v)

        elif isinstance(ov, types.ModuleType):
            continue

        else:
            setattr(obj, k, v)


def config_from_file(file: str) -> pyhocon.ConfigTree:
    return _config_factory.parse_file(file)


def config_from_dict(d: Dict[str, Any]) -> pyhocon.ConfigTree:
    return _config_factory.from_dict(d)


def config_to_dict(config: pyhocon.ConfigTree) -> OrderedDict:
    return config.as_plain_ordered_dict()


def config_merge(config1: pyhocon.ConfigTree, config2: pyhocon.ConfigTree) -> pyhocon.ConfigTree:
    return pyhocon.ConfigTree.merge_configs(config1, config2)




from __future__ import annotations

import asyncio
import logging
import time

from typing import Dict, List, Optional

from qtoggleserver.conf import settings
from qtoggleserver.core import events as core_events
from qtoggleserver.utils import logging as logging_utils


logger = logging.getLogger(__name__)

_sessions_by_id: Dict[str, Session] = {}
_sessions_event_handler: Optional[SessionsEventHandler] = None


class Session(logging_utils.LoggableMixin):
    def __init__(self, session_id: str) -> None:
        logging_utils.LoggableMixin.__init__(self, session_id, logger)

        self.id: str = session_id
        self.accessed: int = 0
        self.timeout: int = 0
        self.access_level: int = 0
        self.future: Optional[asyncio.Future] = None
        self.queue: List[core_events.Event] = []

    def reset_and_wait(self, timeout: int, access_level: int) -> asyncio.Future:
        self.debug('resetting (timeout=%s, access_level=%s)', timeout, access_level)

        if self.future:
            self.debug('already has a listening connection, responding')
            self.respond()

        future = asyncio.get_running_loop().create_future()

        self.accessed = time.time()
        self.timeout = timeout
        self.access_level = access_level
        self.future = future

        if self.queue:
            self.debug('has queued events, responding right away')
            self.respond()

        return future

    def is_empty(self) -> bool:
        return len(self.queue) == 0

    def is_active(self) -> bool:
        return self.future is not None

    def respond(self) -> None:
        events = list(self.queue)
        self.queue = []
        if not self.future:
            return

        self.debug('serving %d events', len(events))
        self.future.set_result(reversed(events))
        self.future = None

    def push(self, event: core_events.Event) -> None:
        # Deduplicate events
        while True:
            duplicates = [e for e in self.queue if event.is_duplicate(e)]
            if not duplicates:
                break

            for d in duplicates:
                self.queue.remove(d)
                self.debug('dropping duplicate event %s', d)

        # Ensure max queue size
        while len(self.queue) >= settings.core.event_queue_size:
            self.warning('queue full, dropping oldest event')
            self.queue.pop()

        self.queue.insert(0, event)

    def __str__(self) -> str:
        return f'session {self.id}'


class SessionsEventHandler(core_events.Handler):
    FIRE_AND_FORGET = False

    def __init__(self, sessions_by_id: Dict[str, Session]) -> None:
        self._sessions_by_id: Dict[str, Session] = sessions_by_id

        super().__init__()

    async def handle_event(self, event: core_events.Event) -> None:
        for session in self._sessions_by_id.values():
            if session.access_level < event.REQUIRED_ACCESS:
                continue

            session.push(event)


def get(session_id: str) -> Session:
    session = _sessions_by_id.get(session_id)
    if not session:
        session = Session(session_id)
        _sessions_by_id[session_id] = session
        session.debug('created')

    return session


def update() -> None:
    now = time.time()
    for session_id, session in list(_sessions_by_id.items()):
        if not session.is_empty() and session.is_active():
            session.respond()
            continue

        if now - session.accessed > session.timeout:
            if session.is_active():
                session.debug('keep-alive')
                session.respond()

            else:
                session.debug('expired')
                _sessions_by_id.pop(session_id)


async def init() -> None:
    global _sessions_event_handler

    _sessions_event_handler = SessionsEventHandler(_sessions_by_id)
    core_events.register_handler(_sessions_event_handler)


async def cleanup() -> None:
    pass




from __future__ import annotations

import functools
import logging

from typing import Any, Callable, Dict

from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.typing import GenericJSONDict


API_VERSION = '1.0'

ACCESS_LEVEL_ADMIN = 30
ACCESS_LEVEL_NORMAL = 20
ACCESS_LEVEL_VIEWONLY = 10
ACCESS_LEVEL_NONE = 0

ACCESS_LEVEL_MAPPING = {
    ACCESS_LEVEL_ADMIN: 'admin',
    ACCESS_LEVEL_NORMAL: 'normal',
    ACCESS_LEVEL_VIEWONLY: 'viewonly',
    ACCESS_LEVEL_NONE: 'none',
    'admin': ACCESS_LEVEL_ADMIN,
    'normal': ACCESS_LEVEL_NORMAL,
    'viewonly': ACCESS_LEVEL_VIEWONLY,
    'none': ACCESS_LEVEL_NONE
}

logger = logging.getLogger(__name__)


class APIError(Exception):
    def __init__(self, status: int, code: str, **params) -> None:
        self.status: int = status
        self.code: str = code
        self.params: dict = params

        super().__init__(code)

    @staticmethod
    def from_http_error(http_error: core_responses.HTTPError) -> APIError:
        return APIError(http_error.status, http_error.code, **http_error.params)

    def to_json(self) -> GenericJSONDict:
        return dict(error=self.code, **self.params)


class APIRequest:
    def __init__(self, handler: APIHandler) -> None:
        self.handler: APIHandler = handler

    @property
    def access_level(self) -> int:
        return self.handler.access_level

    @property
    def username(self) -> str:
        return self.handler.username

    @property
    def method(self) -> str:
        return self.handler.request.method

    @property
    def path(self) -> str:
        return self.handler.request.path

    @property
    def query_arguments(self) -> Dict[str, str]:
        return {k: self.handler.decode_argument(v[0]) for k, v in self.handler.request.query_arguments.items()}

    @property
    def headers(self) -> Dict[str, str]:
        return self.handler.request.headers

    @property
    def body(self) -> bytes:
        return self.handler.request.body


def api_call(access_level: int = ACCESS_LEVEL_NONE) -> Callable:
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(request_handler: APIHandler, *args, **kwargs) -> Any:
            logger.debug('executing API call "%s"', func.__name__)

            if request_handler.access_level < access_level:
                if request_handler.access_level == ACCESS_LEVEL_NONE:  # Indicates missing or invalid auth data
                    raise APIError(401, 'authentication-required')

                else:
                    raise APIError(403, 'forbidden', required_level=ACCESS_LEVEL_MAPPING.get(access_level))

            request = APIRequest(request_handler)

            return func(request, *args, **kwargs)

        return wrapper

    return decorator


# Import this here to prevent errors due to circular imports
from qtoggleserver.web.handlers import APIHandler  # noqa: E402




import asyncio
import datetime
import logging

from typing import List, Optional

from .client import APClient
from .dnsmasq import DNSMasq
from .exceptions import APException
from .hostapd import HostAPD


logger = logging.getLogger(__name__)

_hostapd: Optional[HostAPD] = None
_dnsmasq: Optional[DNSMasq] = None


async def update() -> None:
    global _hostapd
    global _dnsmasq

    if _hostapd and not _hostapd.is_alive():
        logger.warning('hostapd died unexpectedly')
        await _hostapd.stop()
        _hostapd = None

    if _dnsmasq and not _dnsmasq.is_alive():
        logger.warning('dnsmasq died unexpectedly')
        await _dnsmasq.stop()
        _dnsmasq = None

    if None in [_hostapd, _dnsmasq]:
        # Ensure both daemons are stopped
        if _hostapd:
            logger.warning('dnsmasq is stopped but hostapd is running')
            await _hostapd.stop()
            _hostapd = None

        if _dnsmasq:
            logger.warning('hostapd is stopped but dnsmasq is running')
            await _dnsmasq.stop()
            _dnsmasq = None


def is_running() -> bool:
    return (_hostapd is not None) or (_dnsmasq is not None)


def start(
    interface: str,
    ssid: str,
    psk: Optional[str],
    own_ip: str,
    mask_len: int,
    start_ip: str,
    stop_ip: str,
    hostapd_binary: Optional[str] = None,
    hostapd_cli_binary: Optional[str] = None,
    dnsmasq_binary: Optional[str] = None,
    hostapd_log: Optional[str] = None,
    dnsmasq_log: Optional[str] = None
) -> None:

    global _hostapd
    global _dnsmasq

    logger.debug('starting')

    if is_running():
        raise APException('AP already started')

    if HostAPD.is_already_running():
        raise APException('The hostapd daemon is already running')
    if DNSMasq.is_already_running():
        raise APException('The dnsmasq daemon is already running')

    _hostapd = HostAPD(
        interface=interface,
        ssid=ssid,
        psk=psk,
        hostapd_binary=hostapd_binary,
        hostapd_cli_binary=hostapd_cli_binary,
        hostapd_log=hostapd_log
    )

    _dnsmasq = DNSMasq(
        interface=interface,
        own_ip=own_ip,
        mask_len=mask_len,
        start_ip=start_ip,
        stop_ip=stop_ip,
        dnsmasq_binary=dnsmasq_binary,
        dnsmasq_log=dnsmasq_log
    )

    _dnsmasq.start()
    _hostapd.start()

    logger.debug('started')


async def stop() -> None:
    global _hostapd
    global _dnsmasq

    logger.debug('stopping')

    if not is_running():
        raise APException('AP not started')

    await _hostapd.stop()
    await _dnsmasq.stop()

    _hostapd = None
    _dnsmasq = None

    logger.debug('stopped')


def get_clients() -> List[APClient]:
    if not is_running():
        raise APException('AP not started')

    leases = _dnsmasq.get_leases()

    return [APClient(
        mac_address=lease['mac_address'].upper(),
        ip_address=lease['ip_address'],
        hostname=lease['hostname'],
        moment=datetime.datetime.utcfromtimestamp(lease['timestamp'])
    ) for lease in leases]




import sys

from typing import Any


def load_attr(attr_path: str) -> Any:
    m, attr = attr_path.rsplit('.', 1)

    try:
        __import__(m)
        mod = sys.modules[m]

    except ImportError as e:
        raise Exception(f'Error importing {attr_path}: {e}') from e

    try:
        attr = getattr(mod, attr)

    except AttributeError as e:
        raise Exception(f'Error importing {attr_path}: {e}') from e

    return attr




import logging
import os

import psutil

from qtoggleserver.conf import settings

from . import battery
from . import date
from . import dns
from . import fwupdate
from . import net
from . import storage
from . import temperature


logger = logging.getLogger(__name__)


def reboot() -> None:
    logger.debug('rebooting')
    if not settings.debug:
        os.system('reboot')


def uptime() -> int:
    with open('/proc/uptime', 'r') as f:
        line = f.readlines()[0]
        return int(float(line.split()[0]))


def get_cpu_usage() -> int:
    return int(psutil.cpu_percent())


def get_mem_usage() -> int:
    vm = psutil.virtual_memory()
    return int(100 * (1 - vm.available / vm.total))


async def init() -> None:
    logger.debug('initializing DNS')
    await dns.init()


async def cleanup() -> None:
    logger.debug('cleaning up DNS')
    await dns.cleanup()




import re

from typing import Any, Callable, Optional, Tuple, Union

import jsonschema

from qtoggleserver.core.typing import GenericJSONDict

from . import APIError


POST_RESET = {
    "type": "object",
    "properties": {
        "factory": {
            "type": "boolean"
        }
    },
    "additionalProperties": False,
}

PATCH_FIRMWARE = {
    "type": "object",
    "properties": {
        "version": {
            "type": "string",
            "maxLength": 256
        },
        "url": {
            "type": "string",
            "maxLength": 256
        }
    },
    "additionalProperties": False,
    "anyOf": [
        {
            "required": ["version"]
        },
        {
            "required": ["url"]
        }
    ]
}

POST_PORTS = {
    "type": "object",
    "properties": {
        "id": {
            "type": "string",
            "pattern": "^[a-zA-Z_][a-zA-Z0-9_.-]{0,63}$"
        },
        "type": {
            "enum": ["boolean", "number"]
        },
        "min": {
            "type": "number"
        },
        "max": {
            "type": "number"
        },
        "integer": {
            "type": "boolean"
        },
        "step": {
            "type": "number"
        },
        "choices": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "value": {
                        "oneOf": [
                            {"type": "boolean"},
                            {"type": "number"}
                        ]
                    },
                    "display_name": {
                        "type": "string",
                        "maxLength": 64
                    }
                },
                "required": ["value"],
                "additionalProperties": False,
            },
            "minItems": 2,
            "maxItems": 256
        },
    },
    "additionalProperties": False,
    "required": [
        "id",
        "type"
    ]
}

PATCH_PORT_SEQUENCE = {
    "type": "object",
    "properties": {
        "values": {
            "type": "array",
            "items": {
                "oneOf": [
                    {"type": "boolean"},
                    {"type": "number"}
                ]
            },
            "minItems": 0,
            "maxItems": 256
        },
        "delays": {
            "type": "array",
            "items": {
                "type": "integer",
                "min": 0,
                "max": 60000
            },
            "minItems": 0,
            "maxItems": 256
        },
        "repeat": {
            "type": "integer",
            "min": 0,
            "max": 65535
        }
    },
    "additionalProperties": False,
    "required": [
        "values",
        "delays",
        "repeat"
    ]
}

PATCH_WEBHOOKS = {
    "type": "object",
    "properties": {
        "enabled": {
            "type": "boolean"
        },
        "scheme": {
            "enum": ["http", "https"]
        },
        "host": {
            "type": "string",
            "maxLength": 256
        },
        "port": {
            "type": "integer",
            "min": 0,
            "max": 65535
        },
        "path": {
            "type": "string",
            "maxLength": 256
        },
        "timeout": {
            "type": "integer",
            "minimum": 1,
            "maximum": 3600
        },
        "retries": {
            "type": "integer",
            "minimum": 0,
            "maximum": 10
        }
    },
    "additionalProperties": False,
    "required": [
        "enabled",
        "scheme",
        "host",
        "port",
        "path",
        "timeout",
        "retries"
    ]
}

PATCH_REVERSE = {
    "type": "object",
    "properties": {
        "enabled": {
            "type": "boolean"
        },
        "scheme": {
            "enum": ["http", "https"]
        },
        "host": {
            "type": "string",
            "maxLength": 256
        },
        "port": {
            "type": "integer",
            "min": 0,
            "max": 65535
        },
        "path": {
            "type": "string",
            "maxLength": 256
        },
        "device_id": {
            "type": "string",
            "pattern": "^[a-zA-Z0-9]{1,32}"
        },
        "password": {
            "type": "string",
            "maxLength": 32
        },
        "timeout": {
            "type": "integer",
            "minimum": 1,
            "maximum": 3600
        },
    },
    "additionalProperties": False,
    "required": [
        "enabled",
        "scheme",
        "host",
        "port",
        "path"
    ]
}


def _validate_schema(json: Any, schema: GenericJSONDict) -> Optional[Tuple[str, Optional[str]]]:
    try:
        jsonschema.Draft4Validator(schema=schema).validate(json)
        return None

    except jsonschema.ValidationError as e:
        try:
            field = e.path.pop()
            return 'invalid', field

        except Exception:
            try:
                field = re.match(r'\'(\w+)\' is a required property', e.message).group(1)
                return 'missing', field

            except Exception:
                try:
                    field = re.match(r'Additional properties are not allowed \(u?\'(\w+)\'.*', e.message).group(1)
                    return 'unexpected', field

                except Exception:
                    return 'invalid', None


def validate(
    params: Any, schema: GenericJSONDict,
    invalid_field_code: Union[str, Callable] = 'invalid-field',
    unexpected_field_code: Union[str, Callable] = 'invalid-request',
    invalid_request_code: str = 'invalid-request',
    field_name: str = 'field'
) -> None:

    validation_error = _validate_schema(params, schema)
    if validation_error:
        error, field = validation_error
        if error == 'invalid':
            if field:
                if callable(invalid_field_code):
                    invalid_field_code = invalid_field_code(field)

                raise APIError(400, invalid_field_code, **{field_name: field})

            else:
                raise APIError(400, invalid_request_code)

        elif error == 'missing':
            raise APIError(400, 'missing-field', **{field_name: field})

        elif error == 'unexpected':
            if callable(unexpected_field_code):
                unexpected_field_code = unexpected_field_code(field)

            raise APIError(400, unexpected_field_code, **{field_name: field})

        else:
            raise APIError(400, 'invalid-request')




import asyncio
import copy
import re
import time

from typing import Any, List, Optional, Set, Tuple

from qtoggleserver.conf import settings
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.typing import Attribute, Attributes, AttributeDefinitions, GenericJSONDict
from qtoggleserver.core.typing import NullablePortValue, PortValue

from . import exceptions


_DEVICE_EXPRESSION_RE = re.compile(r'^(device_)+expression$')


# We can't use proper type annotations for slaves in this module because that would create unsolvable circular imports.
# Therefore we use "Any" type annotation for Slave instances.


class SlavePort(core_ports.BasePort):
    PERSIST_COLLECTION = 'slave_ports'

    _DEVICE_EXPRESSION_ATTRDEF = {
        'type': 'string',
        'modifiable': True,
        'max': 1024
    }

    _LAST_SYNC_ATTRDEF = {
        'type': 'number',
        'modifiable': False
    }

    _EXPIRES_ATTRDEF = {
        'type': 'number',
        'modifiable': True,
        'min': 0,
        'max': 2147483647
    }

    _PROVISIONING_ATTRDEF = {
        'type': ['string'],
        'modifiable': False
    }

    def __init__(self, slave: Any, attrs: Attributes) -> None:
        from .devices import Slave  # We need this import here just for Slave type annotation

        self._slave: Slave = slave

        self._remote_id: str = attrs['id']

        # Value cache
        self._cached_value: NullablePortValue = None

        # Attributes cache
        self._cached_attrs: Attributes = {}

        # Names of attributes that will be updated remotely asap
        self._remote_update_pending_attrs: Set[Tuple[str, Attribute]] = set()

        # Tag is kept locally
        self._tag: str = ''

        # Timestamp of the last time we've heard from this port
        self._last_sync: float = 0

        # Number of seconds before port value expires
        self._expires: int = 0

        # Cached expired status
        self._expired: bool = False

        # Names of attributes (including value) that have been changed while device was offline and have to be
        # provisioned later
        self._provisioning: Set[str] = set()

        # Helps managing the triggering of port-update event from non-async methods
        self._trigger_update_task: Optional[asyncio.Task] = None

        port_id = f'{slave.get_name()}.{self._remote_id}'

        super().__init__(port_id)

        self.update_cached_attrs(attrs)

        if self._cached_value is not None:
            # Remote value is supplied in attrs when a new port is added on the slave device

            self._value: NullablePortValue = self._cached_value
            self.update_last_sync()

    def _get_standard_attrdefs(self) -> AttributeDefinitions:
        attrdefs = copy.copy(core_ports.STANDARD_ATTRDEFS)

        # Device_*_expression
        for i in range(1, 10):
            slave_name = (i - 1) * 'device_' + 'expression'
            master_name = i * 'device_' + 'expression'
            if slave_name not in self._cached_attrs:
                break

            attrdefs[master_name] = dict(self._DEVICE_EXPRESSION_ATTRDEF)

        # Various master-specific standard attributes
        attrdefs['last_sync'] = dict(self._LAST_SYNC_ATTRDEF)
        attrdefs['expires'] = dict(self._EXPIRES_ATTRDEF)
        attrdefs['provisioning'] = dict(self._PROVISIONING_ATTRDEF)

        return attrdefs

    STANDARD_ATTRDEFS = property(_get_standard_attrdefs)

    def _get_additional_attrdefs(self) -> AttributeDefinitions:
        return copy.copy(self._cached_attrs.get('definitions', {}))

    ADDITIONAL_ATTRDEFS = property(_get_additional_attrdefs)

    def map_id(self, new_id: str) -> None:
        raise core_ports.PortError('Slave ports cannot be mapped')

    def get_remote_id(self) -> str:
        return self._remote_id

    async def get_attr(self, name: str) -> Optional[Attribute]:
        # id - always use the special slave.id notation
        # tag - always kept on master, ignored on slave
        # expression - kept and managed on master, separate attribute for slave
        # device_expression - mapped to expression on slave
        # online - use slave value unless slave itself offline or port disabled
        # expires - always kept on master, ignored on slave

        if name in ('id', 'tag', 'expression', 'online', 'last_sync', 'expires'):
            return await super().get_attr(name)

        elif _DEVICE_EXPRESSION_RE.match(name):
            return self.get_cached_attr(name[7:])

        value = self._cached_attrs.get(name)
        if value is not None:
            return value

        return await super().get_attr(name)

    async def set_attr(self, name: str, value: Attribute) -> None:
        if name in ('tag', 'expression', 'last_sync', 'expires'):
            # Attributes that always stay locally, on master
            await super().set_attr(name, value)

        else:
            if _DEVICE_EXPRESSION_RE.match(name):
                name = name[7:]

            if self._slave.is_online():
                self._remote_update_pending_attrs.add((name, value))

                # Skip an IO loop iteration, allowing setting multiple attributes in one request
                await asyncio.sleep(0)

                try:
                    await self._update_attrs_remotely()

                except Exception as e:
                    # Map exceptions to specific slave API errors
                    raise exceptions.adapt_api_error(e) from e

            else:  # Offline
                # Allow provisioning for offline devices
                self.debug('marking attribute %s for provisioning', name)
                self._provisioning.add(name)
                self._cached_attrs[name] = value

                await self.trigger_update()

    def get_cached_attr(self, name: str) -> Optional[Attribute]:
        return self._cached_attrs.get(name)

    def get_cached_attrs(self) -> Attributes:
        return dict(self._cached_attrs)

    def update_cached_attrs(self, attrs: Attributes) -> None:
        self._cached_attrs = dict(attrs)

        # Value can be found among attrs but we don't want it as attribute
        if 'value' in attrs:
            self._cached_value = self._cached_attrs.pop('value')

        self.invalidate_attrs()
        self.invalidate_attrdefs()

    async def update_enabled(self) -> None:
        if self._cached_attrs.get('enabled') and not self.is_enabled():
            await self.enable()

        elif not self._cached_attrs.get('enabled') and self.is_enabled():
            await self.disable()

    async def _update_attrs_remotely(self) -> None:
        if not self._remote_update_pending_attrs:
            return

        body = {}
        for name, value in self._remote_update_pending_attrs:
            # Transform expressions reference themselves locally via the slave.id identifier; we need to remove the
            # slave name prefix before sending it to the slave
            if name in ('transform_read', 'transform_write'):
                value = value.replace(self._id, self._remote_id)

            body[name] = value

        self._remote_update_pending_attrs = set()

        try:
            await self._slave.api_call('PATCH', f'/ports/{self._remote_id}', body, timeout=settings.slaves.long_timeout)
            self.debug('successfully updated attributes remotely')

        except Exception as e:
            self.debug('failed to update attributes remotely: %s', e)

            raise

    async def is_persisted(self) -> bool:
        # Ports belonging to permanently offline devices should always behave as persisted on master

        if self._slave.is_permanently_offline():
            return True

        return await core_ports.BasePort.is_persisted(self)

    async def attr_is_online(self) -> bool:
        if not self._enabled:
            return False

        if self._expired:
            return False

        if not self._expires and not self._slave.is_online():
            return False

        return self._cached_attrs.get('online', True)

    async def attr_get_provisioning(self) -> List[str]:
        return list(self._provisioning)

    def get_provisioning_attrs(self) -> Attributes:
        provisioning = {}
        for name in self._provisioning:
            value = self._cached_attrs.get(name)
            if value is not None:
                provisioning[name] = value

        return provisioning

    def get_provisioning_value(self) -> Optional[PortValue]:
        if 'value' in self._provisioning:
            return self._cached_value

        return None

    def clear_provisioning(self) -> None:
        self._provisioning = set()

    def get_cached_value(self) -> NullablePortValue:
        return self._cached_value

    def set_cached_value(self, value: PortValue) -> None:
        self._cached_value = value

    async def read_value(self) -> NullablePortValue:
        return self._cached_value

    async def write_value(self, value: PortValue) -> None:
        if self._slave.is_online():
            try:
                await self._slave.api_call(
                    'PATCH',
                    f'/ports/{self._remote_id}/value',
                    value,
                    timeout=settings.slaves.long_timeout
                )

                # Immediately fetch the current port value from slave and update it locally. This ensures that calls to
                # Port.get_value() will return the correct updated value.
                remote_value = await self._slave.api_call('GET', f'/ports/{self._remote_id}/value')
                self.set_cached_value(remote_value)

            except core_responses.HTTPError as e:
                if e.code == 502 and e.code == 'port-error':
                    message = e.params.get('message')
                    if message:
                        raise core_ports.PortError(message)

                    else:
                        raise core_ports.PortError()

                if e.code == 504 and e.code == 'port-timeout':
                    raise core_ports.PortTimeout()

                raise exceptions.adapt_api_error(e) from e

        else:  # Offline
            # Allow provisioning for offline devices
            self.debug('marking value for provisioning')
            self._cached_value = value
            self._provisioning.add('value')
            await self.save()  # Save provisioning value

            # We need to trigger a port-update because our provisioning attribute has changed
            await self.trigger_update()

    async def set_sequence(self, values: List[PortValue], delays: List[int], repeat: int) -> None:
        if not self._slave.is_online():
            raise exceptions.DeviceOffline(self._slave)

        try:
            await self._slave.api_call(
                'PATCH',
                f'/ports/{self._remote_id}/sequence',
                {'values': values, 'delays': delays, 'repeat': repeat}
            )
            self.debug('sequence sent remotely')

        except Exception as e:
            self.error('failed to send sequence remotely: %s', e)

            # Map exceptions to specific slave API errors
            raise exceptions.adapt_api_error(e) from e

    def heart_beat_second(self) -> None:
        was_expired = self._expired
        now_expired = (self._expires > 0) and (time.time() - self._last_sync > self._expires)
        self._expired = now_expired

        if was_expired != now_expired:
            if now_expired:
                self.debug('value expired')

            if not self._trigger_update_task:
                self._trigger_update_task = asyncio.create_task(self.trigger_update())

    async def load_from_data(self, data: GenericJSONDict) -> None:
        # Only consider locally persisted attributes for permanently offline devices. For online devices, we always use
        # fresh attributes received from device.
        attrs = data.get('attrs')
        if attrs and self._slave.is_permanently_offline():
            if 'value' in data:
                attrs['value'] = data['value']

            self.update_cached_attrs(attrs)

        # Attributes that are kept on master
        for attr in ('tag', 'expression', 'last_sync', 'expires'):
            if attr in data:
                await self.set_attr(attr, data[attr])

        self._provisioning = set(data.get('provisioning', []))

        # Enable if enabled remotely
        await self.update_enabled()

    async def prepare_for_save(self) -> GenericJSONDict:
        return {
            'id': self.get_id(),
            'tag': self._tag,
            'expression': str(self._expression or ''),
            'last_sync': self._last_sync,
            'expires': self._expires,
            'provisioning': list(self._provisioning),
            'attrs': self._cached_attrs,
            'value': self._cached_value
        }

    def update_last_sync(self) -> None:
        self._last_sync = int(time.time())

    async def handle_enable(self) -> None:
        # Fetch current port value, but not before slaves are ready. Slaves are't ready during initial loading at
        # startup, at which point we've got a recent values for all slave ports
        if self._slave.is_ready():
            self.debug('fetching port value')

            try:
                value = await self._slave.api_call('GET', f'/ports/{self._remote_id}/value', retry_counter=None)

            except Exception as e:
                self.error('failed to fetch port value: %s', e)

            else:
                if value is not None:
                    self._cached_value = value

    async def trigger_update(self) -> None:
        await super().trigger_update()

        self._trigger_update_task = None

    async def cleanup(self) -> None:
        await super().cleanup()

        if self._trigger_update_task:
            await self._trigger_update_task




from qtoggleserver.core import api as core_api
from qtoggleserver.core import webhooks as core_webhooks
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.typing import GenericJSONDict


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_webhooks(request: core_api.APIRequest) -> GenericJSONDict:
    return core_webhooks.get().to_json()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_webhooks(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    core_api_schema.validate(params, core_api_schema.PATCH_WEBHOOKS)

    try:
        core_webhooks.setup(**params)

    except core_webhooks.InvalidParamError as e:
        raise core_api.APIError(400, 'invalid-field', field=e.param) from e

    core_webhooks.save()




import logging

from typing import List, Optional

from tornado.web import Application, RequestHandler, URLSpec
from qui.web import tornado as qui_tornado

from qtoggleserver.conf import settings
from qtoggleserver.slaves.discover import is_enabled as is_discover_enabled
from qtoggleserver.web import handlers


logger = logging.getLogger(__name__)

_application: Optional[Application] = None


def _log_request(handler: RequestHandler) -> None:
    if handler.get_status() < 400:
        log_method = logger.debug

    elif handler.get_status() < 500:
        log_method = logger.warning

    else:
        log_method = logger.error

    request_time = 1000.0 * handler.request.request_time()
    log_method('%d %s %.2fms', handler.get_status(), handler._request_summary(), request_time)


def _make_routing_table() -> List[URLSpec]:
    handlers_list = []

    # Frontend

    if settings.frontend.enabled:
        handlers_list += qui_tornado.make_routing_table()

        handlers_list += [
            URLSpec(r'^/api/frontend/dashboard/panels/?$', handlers.DashboardPanelsHandler),
            URLSpec(r'^/api/frontend/prefs/?$', handlers.PrefsHandler),
        ]

    handlers_list += [
        # Device management
        URLSpec(r'^/api/device/?$', handlers.DeviceHandler),
        URLSpec(r'^/api/reset/?$', handlers.ResetHandler),
        URLSpec(r'^/api/access/?$', handlers.AccessHandler),

        # Port management
        URLSpec(r'^/api/ports/?$', handlers.PortsHandler),
        URLSpec(r'^/api/ports/(?P<port_id>[A-Za-z0-9_.-]+)/?$', handlers.PortHandler),

        # Port values
        URLSpec(r'^/api/ports/(?P<port_id>[A-Za-z0-9_.-]+)/value/?$', handlers.PortValueHandler),
    ]

    if settings.core.sequences_support:
        handlers_list += [
            URLSpec(r'^/api/ports/(?P<port_id>[A-Za-z0-9_.-]+)/sequence/?$', handlers.PortSequenceHandler)
        ]

    # Firmware

    if settings.system.fwupdate.driver:
        handlers_list += [
            URLSpec(r'^/api/firmware/?$', handlers.FirmwareHandler)
        ]

    # Slave devices management

    if settings.slaves.enabled:
        handlers_list += [
            URLSpec(r'^/api/devices/?$', handlers.SlaveDevicesHandler),
            URLSpec(r'^/api/devices/(?P<name>[A-Za-z0-9_-]+)/?$', handlers.SlaveDeviceHandler),
            URLSpec(r'^/api/devices/(?P<name>[A-Za-z0-9_-]+)/events/?$', handlers.SlaveDeviceEventsHandler),
            URLSpec(r'^/api/devices/(?P<name>[A-Za-z0-9_-]+)/forward/(?P<path>.+)$', handlers.SlaveDeviceForwardHandler)
        ]

        if is_discover_enabled():
            handlers_list += [
                URLSpec(r'^/api/discovered/?$', handlers.DiscoveredHandler),
                URLSpec(r'^/api/discovered/(?P<name>[A-Za-z0-9_-]+)/?$', handlers.DiscoveredDeviceHandler)
            ]

    # Notifications

    if settings.webhooks.enabled:
        handlers_list += [
            URLSpec(r'^/api/webhooks/?$', handlers.WebhooksHandler)
        ]

    if settings.core.listen_support:
        handlers_list += [
            URLSpec(r'^/api/listen/?$', handlers.ListenHandler)
        ]

    # Reverse API calls

    if settings.reverse.enabled:
        handlers_list += [
            URLSpec(r'^/api/reverse/?$', handlers.ReverseHandler)
        ]

    handlers_list += [
        URLSpec(r'^/api/.*$', handlers.NoSuchFunctionHandler)
    ]

    return handlers_list


def get_application() -> Application:
    global _application

    if _application is None:
        _application = Application(
            handlers=_make_routing_table(),
            debug=False,
            compress_response=settings.server.compress_response,
            log_function=_log_request
        )

    return _application




import subprocess

from typing import Dict, Optional

from qtoggleserver.conf import settings
from qtoggleserver.utils.cmd import run_get_cmd, run_set_cmd


class NetError(Exception):
    pass


def has_ip_support() -> bool:
    return bool(settings.system.net.ip.get_cmd and settings.system.net.ip.set_cmd)


def get_ip_config() -> Dict[str, str]:
    return run_get_cmd(
        settings.system.net.ip.get_cmd,
        cmd_name='IP config',
        exc_class=NetError,
        required_fields=[
            'address',
            'netmask',
            'gateway',
            'dns',
            'address_current',
            'netmask_current',
            'gateway_current',
            'dns_current'
        ]
    )


def set_ip_config(address: str, netmask: str, gateway: str, dns: str) -> None:
    run_set_cmd(
        settings.system.net.ip.set_cmd,
        cmd_name='IP config',
        exc_class=NetError,
        address=address,
        netmask=netmask,
        gateway=gateway,
        dns=dns
    )


def has_wifi_support() -> bool:
    return bool(settings.system.net.wifi.get_cmd and settings.system.net.wifi.set_cmd)


def get_wifi_config() -> Dict[str, str]:
    return run_get_cmd(
        settings.system.net.wifi.get_cmd,
        cmd_name='WiFi config',
        log_values=False,
        exc_class=NetError,
        required_fields=['ssid', 'psk', 'bssid', 'bssid_current', 'rssi_current']
    )


def set_wifi_config(ssid: str, psk: str, bssid: str) -> None:
    run_set_cmd(
        settings.system.net.wifi.set_cmd,
        cmd_name='WiFi config',
        exc_class=NetError,
        log_values=False,
        ssid=ssid,
        psk=psk,
        bssid=bssid
    )


def get_default_interface() -> Optional[str]:
    try:
        output = subprocess.check_output(['ip', 'route']).decode().strip()

    except Exception as e:
        raise NetError('Could not determine default route') from e

    lines = output.split('\n')
    if not lines:
        raise NetError('Could not determine default route')

    line = next((line for line in lines if line.count('default via')), None)
    if not line:
        return

    return line.split()[4]  # The 5th part is the interface




import re

from typing import Dict, List, Optional

from qtoggleserver import slaves
from qtoggleserver import system
from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core import device as core_device
from qtoggleserver.core import main
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core import reverse as core_reverse
from qtoggleserver.core import sessions as core_sessions
from qtoggleserver.core import vports as core_vports
from qtoggleserver.core import webhooks as core_webhooks
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.typing import GenericJSONDict


@core_api.api_call(core_api.ACCESS_LEVEL_NONE)
async def get_access(request: core_api.APIRequest, access_level: int) -> Dict[str, str]:
    return {
        'level': core_api.ACCESS_LEVEL_MAPPING[access_level]
    }


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_listen(
    request: core_api.APIRequest,
    session_id: str,
    timeout: Optional[int],
    access_level: int
) -> List[GenericJSONDict]:

    if session_id is None:
        raise core_api.APIError(400, 'missing-field', field='session_id')

    if not re.match('[a-zA-Z0-9]{1,32}', session_id):
        raise core_api.APIError(400, 'invalid-field', field='session_id')

    if timeout is not None:
        try:
            timeout = int(timeout)

        except Exception:
            raise core_api.APIError(400, 'invalid-field', field='timeout') from None

        if timeout < 1 or timeout > 3600:
            raise core_api.APIError(400, 'invalid-field', field='timeout')

    else:
        timeout = 60

    session = core_sessions.get(session_id)
    events = await session.reset_and_wait(timeout, access_level)

    return [await e.to_json() for e in events]


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def post_reset(request: core_api.APIRequest, params: GenericJSONDict) -> None:
    core_api_schema.validate(params, core_api_schema.POST_RESET)

    factory = params.get('factory')

    if factory:
        core_api.logger.info('resetting to factory defaults')

        core_ports.reset()
        core_vports.reset()
        core_device.reset()
        if settings.webhooks.enabled:
            core_webhooks.reset()
        if settings.reverse.enabled:
            core_reverse.reset()
        if settings.slaves.enabled:
            slaves.reset()

    main.loop.call_later(2, system.reboot)




from .apclients import DiscoveredDevice, DiscoverException
from .apclients import discover, get_discovered_devices, configure, finish
from .apclients import init as apclients_init, cleanup as apclients_cleanup
from .apclients import get_interface as get_apclients_interface


def is_enabled() -> bool:
    # Discover mechanism is enabled if AP clients method is enabled;
    # AP clients method is enabled if it has available interface.
    return get_apclients_interface() is not None


async def init() -> None:
    await apclients_init()


async def cleanup() -> None:
    await apclients_cleanup()




from typing import Optional

import psutil

from qtoggleserver.conf import settings
from qtoggleserver.utils.cmd import run_get_cmd


class TemperatureError(Exception):
    pass


def has_temperature_support() -> bool:
    return bool(settings.system.temperature.sensor_name) or bool(settings.system.temperature.get_cmd)


def get_temperature() -> Optional[int]:
    if settings.system.temperature.get_cmd:
        return int(run_get_cmd(
            settings.system.temperature.get_cmd,
            cmd_name='temperature',
            exc_class=TemperatureError,
            required_fields=['value']
        )['value'])

    if settings.system.temperature.sensor_name:
        temperatures = psutil.sensors_temperatures()
        temp_info_list = temperatures[settings.system.temperature.sensor_name]
        temp_info = temp_info_list[settings.system.temperature.sensor_index]

        return int(temp_info.current)




from qtoggleserver import system
from qtoggleserver.core import api as core_api
from qtoggleserver.core import device as core_device
from qtoggleserver.core import main
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core.device import events as core_device_events
from qtoggleserver.core.typing import Attributes


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_device(request: core_api.APIRequest) -> Attributes:
    return core_device_attrs.to_json()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_device(request: core_api.APIRequest, params: Attributes) -> None:
    def unexpected_field_code(field: str) -> str:
        if field in core_device_attrs.get_attrdefs():
            return 'attribute-not-modifiable'

        else:
            return 'no-such-attribute'

    core_api_schema.validate(
        params,
        core_device_attrs.get_schema(),
        unexpected_field_code=unexpected_field_code
    )

    try:
        reboot_required = core_device_attrs.set_attrs(params)

    except core_device_attrs.DeviceAttributeError as e:
        raise core_api.APIError(400, e.error, attribute=e.attribute)

    except Exception as e:
        raise core_api.APIError(500, 'unexpected-error', message=str(e)) from e

    core_device.save()
    await core_device_events.trigger_update()

    if reboot_required:
        main.loop.call_later(2, system.reboot)
